from typing import Callable, Dict, Generic, Hashable, Iterable, Iterator, List, TypeVar, Union

import numpy as np
from nltk.tokenize import word_tokenize

from .func import identity

T = TypeVar("T")


class GenericLabelEncoder:
    """Encodes all hashable objects to integers"""

    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        self.object2idx: Dict[Hashable, int] = {}
        self.idx2object: List[Hashable] = []

    def encode(self, obj: Hashable) -> int:
        """Encode a hashable `obj` to an integer."""

        size = len(self.idx2object)
        idx = self.object2idx.setdefault(obj, size)
        if idx == size:  # inserted new object
            self.idx2object.append(obj)

        return idx

    def encode_batch(self, objs: Iterable[Hashable]) -> Iterator[int]:
        for obj in objs:
            yield self.encode(obj)

    def decode(self, idx: int) -> Hashable:
        """Decode an integer `idx` to the previously encoded object.
        Raises IndexError if the index is out of range.
        """

        return self.idx2object[idx]

    def decode_batch(self, indices: Iterable[int]) -> Iterator[Hashable]:
        for idx in indices:
            yield self.decode(idx)

    def __len__(self) -> int:
        assert len(self.object2idx) == len(self.idx2object)
        return len(self.idx2object)


class BatchLabelEncoder(Generic[T]):
    """Similar to `sklearn.preprocessing.LabelEncoder` but accepts a list of sentences as inputs
    and returns a list of lists of integer labels.
    Warning: Calling `fit` doesn't reset the encoder. To do so call `reset()` explicitly.
    cf. `keras_preprocessing.text.Tokenizer`
    """

    def __init__(self, tokenizer: str) -> None:
        self.tokenizer: Callable[[str], Iterator[T]] = {
            "none": identity,
            "nltk": word_tokenize,
        }[tokenizer]
        self.reset()

    def reset(self) -> None:
        """Initializes all mappings with empty containers."""

        self.token2idx: Dict[T, int] = {}

        self.idx2count: Union[List[int], np.ndarray] = []
        self.idx2token: Union[List[T], np.ndarray] = []

    def finalize(self, vocab_size: int) -> None:
        """This only keeps the most frequent `vocab_size` words in the vocabulary.
        This will change the assigned indices, so if `transform` was used before `finalize`,
        they cannot be transformed back using `inverse_transform`.
        """

        self.idx2count = np.array(self.idx2count)
        self.idx2token = np.array(self.idx2token)
        indices = np.argsort(-self.idx2count)[:vocab_size]

        self.idx2count = self.idx2count[indices]
        self.idx2token = self.idx2token[indices]

        self.token2idx = {token: i for i, token in enumerate(self.idx2token)}

    def partial_fit_single(self, token: T) -> int:
        vocab_size = self.num_labels
        idx = self.token2idx.setdefault(token, vocab_size)
        if idx == vocab_size:  # inserted new token
            self.idx2count.append(1)
            self.idx2token.append(token)
        else:
            self.idx2count[idx] += 1

        return idx

    def partial_fit(self, sentence: str) -> None:
        for token in self.tokenizer(sentence):
            self.partial_fit_single(token)

    def partial_fit_batch(self, sentences: Iterable[str]) -> None:
        for sentence in sentences:
            self.partial_fit(sentence)

    def fit_single(self, token: T) -> None:
        self.partial_fit_single(token)

    def fit(self, sentence: str) -> None:
        self.partial_fit(sentence)

    def fit_batch(self, sentences: Iterable[str]) -> None:
        self.partial_fit_batch(sentences)

    def transform_single(self, token: T) -> int:
        """Unknown labels are skipped."""

        return self.token2idx[token]

    def transform(self, sentence: str, ignore: bool = True) -> Iterator[int]:
        for token in self.tokenizer(sentence):
            try:
                yield self.transform_single(token)
            except KeyError:
                if ignore:
                    pass
                else:
                    raise

    def transform_batch(self, sentences: Iterable[str], ignore: bool = True) -> Iterator[List[int]]:
        """Unknown labels are skipped."""

        for sentence in sentences:
            yield list(self.transform(sentence, ignore))

    def fit_transform_single(self, token: T) -> int:
        return self.partial_fit_single(token)

    def fit_transform(self, sentences: str) -> Iterator[int]:
        self.partial_fit(sentences)
        return self.transform(sentences)

    def fit_transform_batch(self, sentences: Iterable[str]) -> Iterator[List[int]]:
        self.partial_fit_batch(sentences)
        return self.transform_batch(sentences)

    def inverse_transform_single(self, idx: int) -> T:
        return self.idx2token[idx]

    def inverse_transform(self, indices: Iterable[int]) -> Iterator[T]:
        for idx in indices:
            yield self.inverse_transform_single(idx)

    def inverse_transform_batch(self, list_of_indices: Iterable[Iterable[int]]) -> Iterator[List[T]]:
        for indices in list_of_indices:
            yield list(self.inverse_transform(indices))

    @property
    def num_labels(self) -> int:
        assert len(self.token2idx) == len(self.idx2count) == len(self.idx2token)
        return len(self.idx2token)



import sys

if sys.platform != "win32":
    raise ImportError("win submodule is only available on Windows")

import logging
import os
import re
from argparse import ArgumentTypeError
from ctypes import POINTER, Structure, addressof, byref, cast, create_unicode_buffer, sizeof, string_at
from ctypes.wintypes import BYTE, DWORD, HANDLE, PBYTE, ULONG
from enum import Enum, IntEnum, IntFlag
from textwrap import indent
from typing import Any, Iterator, List, Optional, Tuple, Union

from cwinsdk import struct2dict
from cwinsdk.shared import devpkey, devpropdef, diskguid, ntddscsi, ntddstor, nvme, scsi, srb, winerror
from cwinsdk.shared.basetsd import UINT32
from cwinsdk.shared.guiddef import GUID
from cwinsdk.um import setupapi, winioctl
from cwinsdk.um.fileapi import (
    OPEN_EXISTING,
    CreateFileW,
    FindFirstVolumeW,
    FindNextVolumeW,
    FindVolumeClose,
    GetLogicalDriveStringsW,
    GetVolumeNameForVolumeMountPointW,
    GetVolumePathNamesForVolumeNameW,
    QueryDosDeviceW,
)
from cwinsdk.um.handleapi import CloseHandle
from cwinsdk.um.ioapiset import CancelIo, DeviceIoControl, GetOverlappedResultEx
from cwinsdk.um.minwinbase import OVERLAPPED
from cwinsdk.um.synchapi import CreateEventW
from cwinsdk.um.winbase import (
    FILE_FLAG_OVERLAPPED,
    FindFirstVolumeMountPointW,
    FindNextVolumeMountPointW,
    FindVolumeMountPointClose,
)
from cwinsdk.um.winnt import FILE_SHARE_READ, FILE_SHARE_WRITE
from typing_extensions import Self

from ..exceptions import assert_true
from .handle import WindowsHandle, _mode2access

assert sizeof(winioctl.STORAGE_PROTOCOL_SPECIFIC_DATA) == 40

logger = logging.getLogger(__name__)

win32_drive_letter_p = re.compile(r"^\\\\[\.\?]\\[A-Z]:$", flags=re.IGNORECASE)
physical_drive_p = re.compile(r"^\\\\[\.\?]\\PHYSICALDRIVE[0-9]+$")
volume_guid_path_p = re.compile(
    r"^\\\\\?\\Volume\{[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}\}\\$", flags=re.IGNORECASE
)
logical_drive_p = re.compile(r"^[A-Z]:\\$", flags=re.IGNORECASE)


class SrbType(IntEnum):
    SCSI_REQUEST_BLOCK = srb.SRB_TYPE_SCSI_REQUEST_BLOCK
    STORAGE_REQUEST_BLOCK = srb.SRB_TYPE_STORAGE_REQUEST_BLOCK
    NVME_REQUEST_BLOCK = srb.SRB_TYPE_NVME_REQUEST_BLOCK


class IsDirtyFlags(IntFlag):
    VOLUME_IS_CLEAN = 0
    VOLUME_IS_DIRTY = winioctl.VOLUME_IS_DIRTY
    VOLUME_UPGRADE_SCHEDULED = winioctl.VOLUME_UPGRADE_SCHEDULED
    VOLUME_SESSION_OPEN = winioctl.VOLUME_SESSION_OPEN


class StorageAddressType(IntEnum):
    BTL8 = srb.STORAGE_ADDRESS_TYPE_BTL8
    NVME = srb.STORAGE_ADDRESS_TYPE_NVME


class ScsiDriveType(IntEnum):
    DIRECT_ACCESS_DEVICE = scsi.DIRECT_ACCESS_DEVICE
    SEQUENTIAL_ACCESS_DEVICE = scsi.SEQUENTIAL_ACCESS_DEVICE
    PRINTER_DEVICE = scsi.PRINTER_DEVICE
    PROCESSOR_DEVICE = scsi.PROCESSOR_DEVICE
    WRITE_ONCE_READ_MULTIPLE_DEVICE = scsi.WRITE_ONCE_READ_MULTIPLE_DEVICE
    READ_ONLY_DIRECT_ACCESS_DEVICE = scsi.READ_ONLY_DIRECT_ACCESS_DEVICE
    SCANNER_DEVICE = scsi.SCANNER_DEVICE
    OPTICAL_DEVICE = scsi.OPTICAL_DEVICE
    MEDIUM_CHANGER = scsi.MEDIUM_CHANGER
    COMMUNICATION_DEVICE = scsi.COMMUNICATION_DEVICE
    ARRAY_CONTROLLER_DEVICE = scsi.ARRAY_CONTROLLER_DEVICE
    SCSI_ENCLOSURE_DEVICE = scsi.SCSI_ENCLOSURE_DEVICE
    REDUCED_BLOCK_DEVICE = scsi.REDUCED_BLOCK_DEVICE
    OPTICAL_CARD_READER_WRITER_DEVICE = scsi.OPTICAL_CARD_READER_WRITER_DEVICE
    BRIDGE_CONTROLLER_DEVICE = scsi.BRIDGE_CONTROLLER_DEVICE
    OBJECT_BASED_STORAGE_DEVICE = scsi.OBJECT_BASED_STORAGE_DEVICE
    HOST_MANAGED_ZONED_BLOCK_DEVICE = scsi.HOST_MANAGED_ZONED_BLOCK_DEVICE
    UNKNOWN_OR_NO_DEVICE = scsi.UNKNOWN_OR_NO_DEVICE
    LOGICAL_UNIT_NOT_PRESENT_DEVICE = scsi.LOGICAL_UNIT_NOT_PRESENT_DEVICE


class FILESYSTEM_STATISTICS_TYPE(IntEnum):
    NTFS = winioctl.FILESYSTEM_STATISTICS_TYPE_NTFS
    FAT = winioctl.FILESYSTEM_STATISTICS_TYPE_FAT
    EXFAT = winioctl.FILESYSTEM_STATISTICS_TYPE_EXFAT
    REFS = winioctl.FILESYSTEM_STATISTICS_TYPE_REFS


class PartitionInt(IntEnum):
    ENTRY_UNUSED = winioctl.PARTITION_ENTRY_UNUSED
    IFS = winioctl.PARTITION_IFS
    FAT32 = winioctl.PARTITION_FAT32
    MSFT_RECOVERY = winioctl.PARTITION_MSFT_RECOVERY


class PartitionGuid(Enum):
    BASIC_DATA = diskguid.PARTITION_BASIC_DATA_GUID
    BSP = diskguid.PARTITION_BSP_GUID
    CLUSTER = diskguid.PARTITION_CLUSTER_GUID
    DPP = diskguid.PARTITION_DPP_GUID
    ENTRY_UNUSED = diskguid.PARTITION_ENTRY_UNUSED_GUID
    LDM_DATA = diskguid.PARTITION_LDM_DATA_GUID
    LDM_METADATA = diskguid.PARTITION_LDM_METADATA_GUID
    LEGACY_BL = diskguid.PARTITION_LEGACY_BL_GUID
    LEGACY_BL_BACKUP = diskguid.PARTITION_LEGACY_BL_GUID_BACKUP
    MAIN_OS = diskguid.PARTITION_MAIN_OS_GUID
    MSFT_RECOVERY = diskguid.PARTITION_MSFT_RECOVERY_GUID
    MSFT_RESERVED = diskguid.PARTITION_MSFT_RESERVED_GUID
    MSFT_SNAPSHOT = diskguid.PARTITION_MSFT_SNAPSHOT_GUID
    OS_DATA = diskguid.PARTITION_OS_DATA_GUID
    PATCH = diskguid.PARTITION_PATCH_GUID
    PRE_INSTALLED = diskguid.PARTITION_PRE_INSTALLED_GUID
    SBL_CACHE_SSD = diskguid.PARTITION_SBL_CACHE_SSD_GUID
    SBL_CACHE_SSD_RESERVED = diskguid.PARTITION_SBL_CACHE_SSD_RESERVED_GUID
    SBL_CACHE_HDD = diskguid.PARTITION_SBL_CACHE_HDD_GUID
    SERVICING_FILES = diskguid.PARTITION_SERVICING_FILES_GUID
    SERVICING_METADATA = diskguid.PARTITION_SERVICING_METADATA_GUID
    SERVICING_RESERVE = diskguid.PARTITION_SERVICING_RESERVE_GUID
    SERVICING_STAGING_ROOT = diskguid.PARTITION_SERVICING_STAGING_ROOT_GUID
    SPACES = diskguid.PARTITION_SPACES_GUID
    SPACES_DATA = diskguid.PARTITION_SPACES_DATA_GUID
    SYSTEM = diskguid.PARTITION_SYSTEM_GUID
    WINDOWS_SYSTEM = diskguid.PARTITION_WINDOWS_SYSTEM_GUID


class DEVICE_TYPE(IntEnum):
    BEEP = winioctl.FILE_DEVICE_BEEP
    CD_ROM = winioctl.FILE_DEVICE_CD_ROM
    CD_ROM_FILE_SYSTEM = winioctl.FILE_DEVICE_CD_ROM_FILE_SYSTEM
    CONTROLLER = winioctl.FILE_DEVICE_CONTROLLER
    DATALINK = winioctl.FILE_DEVICE_DATALINK
    DFS = winioctl.FILE_DEVICE_DFS
    DISK = winioctl.FILE_DEVICE_DISK
    DISK_FILE_SYSTEM = winioctl.FILE_DEVICE_DISK_FILE_SYSTEM
    FILE_SYSTEM = winioctl.FILE_DEVICE_FILE_SYSTEM
    INPORT_PORT = winioctl.FILE_DEVICE_INPORT_PORT
    KEYBOARD = winioctl.FILE_DEVICE_KEYBOARD
    MAILSLOT = winioctl.FILE_DEVICE_MAILSLOT
    MIDI_IN = winioctl.FILE_DEVICE_MIDI_IN
    MIDI_OUT = winioctl.FILE_DEVICE_MIDI_OUT
    MOUSE = winioctl.FILE_DEVICE_MOUSE
    MULTI_UNC_PROVIDER = winioctl.FILE_DEVICE_MULTI_UNC_PROVIDER
    NAMED_PIPE = winioctl.FILE_DEVICE_NAMED_PIPE
    NETWORK = winioctl.FILE_DEVICE_NETWORK
    NETWORK_BROWSER = winioctl.FILE_DEVICE_NETWORK_BROWSER
    NETWORK_FILE_SYSTEM = winioctl.FILE_DEVICE_NETWORK_FILE_SYSTEM
    NULL = winioctl.FILE_DEVICE_NULL
    PARALLEL_PORT = winioctl.FILE_DEVICE_PARALLEL_PORT
    PHYSICAL_NETCARD = winioctl.FILE_DEVICE_PHYSICAL_NETCARD
    PRINTER = winioctl.FILE_DEVICE_PRINTER
    SCANNER = winioctl.FILE_DEVICE_SCANNER
    SERIAL_MOUSE_PORT = winioctl.FILE_DEVICE_SERIAL_MOUSE_PORT
    SERIAL_PORT = winioctl.FILE_DEVICE_SERIAL_PORT
    SCREEN = winioctl.FILE_DEVICE_SCREEN
    SOUND = winioctl.FILE_DEVICE_SOUND
    STREAMS = winioctl.FILE_DEVICE_STREAMS
    TAPE = winioctl.FILE_DEVICE_TAPE
    TAPE_FILE_SYSTEM = winioctl.FILE_DEVICE_TAPE_FILE_SYSTEM
    TRANSPORT = winioctl.FILE_DEVICE_TRANSPORT
    UNKNOWN = winioctl.FILE_DEVICE_UNKNOWN
    VIDEO = winioctl.FILE_DEVICE_VIDEO
    VIRTUAL_DISK = winioctl.FILE_DEVICE_VIRTUAL_DISK
    WAVE_IN = winioctl.FILE_DEVICE_WAVE_IN
    WAVE_OUT = winioctl.FILE_DEVICE_WAVE_OUT
    FILE_DEVICE_8042_PORT = 0x00000027
    NETWORK_REDIRECTOR = winioctl.FILE_DEVICE_NETWORK_REDIRECTOR
    BATTERY = winioctl.FILE_DEVICE_BATTERY
    BUS_EXTENDER = winioctl.FILE_DEVICE_BUS_EXTENDER
    MODEM = winioctl.FILE_DEVICE_MODEM
    VDM = winioctl.FILE_DEVICE_VDM
    MASS_STORAGE = winioctl.FILE_DEVICE_MASS_STORAGE
    SMB = winioctl.FILE_DEVICE_SMB
    KS = winioctl.FILE_DEVICE_KS
    CHANGER = winioctl.FILE_DEVICE_CHANGER
    SMARTCARD = winioctl.FILE_DEVICE_SMARTCARD
    ACPI = winioctl.FILE_DEVICE_ACPI
    DVD = winioctl.FILE_DEVICE_DVD
    FULLSCREEN_VIDEO = winioctl.FILE_DEVICE_FULLSCREEN_VIDEO
    DFS_FILE_SYSTEM = winioctl.FILE_DEVICE_DFS_FILE_SYSTEM
    DFS_VOLUME = winioctl.FILE_DEVICE_DFS_VOLUME
    SERENUM = winioctl.FILE_DEVICE_SERENUM
    TERMSRV = winioctl.FILE_DEVICE_TERMSRV
    KSEC = winioctl.FILE_DEVICE_KSEC
    FIPS = winioctl.FILE_DEVICE_FIPS
    INFINIBAND = winioctl.FILE_DEVICE_INFINIBAND
    VMBUS = winioctl.FILE_DEVICE_VMBUS
    CRYPT_PROVIDER = winioctl.FILE_DEVICE_CRYPT_PROVIDER
    WPD = winioctl.FILE_DEVICE_WPD
    BLUETOOTH = winioctl.FILE_DEVICE_BLUETOOTH
    MT_COMPOSITE = winioctl.FILE_DEVICE_MT_COMPOSITE
    MT_TRANSPORT = winioctl.FILE_DEVICE_MT_TRANSPORT
    BIOMETRIC = winioctl.FILE_DEVICE_BIOMETRIC
    PMI = winioctl.FILE_DEVICE_PMI
    EHSTOR = winioctl.FILE_DEVICE_EHSTOR
    DEVAPI = winioctl.FILE_DEVICE_DEVAPI
    GPIO = winioctl.FILE_DEVICE_GPIO
    USBEX = winioctl.FILE_DEVICE_USBEX
    CONSOLE = winioctl.FILE_DEVICE_CONSOLE
    NFP = winioctl.FILE_DEVICE_NFP
    SYSENV = winioctl.FILE_DEVICE_SYSENV
    VIRTUAL_BLOCK = winioctl.FILE_DEVICE_VIRTUAL_BLOCK
    POINT_OF_SERVICE = winioctl.FILE_DEVICE_POINT_OF_SERVICE
    STORAGE_REPLICATION = winioctl.FILE_DEVICE_STORAGE_REPLICATION
    TRUST_ENV = winioctl.FILE_DEVICE_TRUST_ENV
    UCM = winioctl.FILE_DEVICE_UCM
    UCMTCPCI = winioctl.FILE_DEVICE_UCMTCPCI
    PERSISTENT_MEMORY = winioctl.FILE_DEVICE_PERSISTENT_MEMORY
    NVDIMM = winioctl.FILE_DEVICE_NVDIMM
    HOLOGRAPHIC = winioctl.FILE_DEVICE_HOLOGRAPHIC
    SDFXHCI = winioctl.FILE_DEVICE_SDFXHCI
    UCMUCSI = winioctl.FILE_DEVICE_UCMUCSI


def is_volume_guid_path(s: str) -> bool:
    return volume_guid_path_p.match(s) is not None


def is_volume_guid_path_arg(s: str) -> str:
    if is_volume_guid_path(s):
        return s
    raise ArgumentTypeError("\\\\?\\Volume{GUID}\\")


def is_logical_drive(s: str) -> bool:
    return logical_drive_p.match(s) is not None


def is_logical_drive_arg(s: str) -> str:
    if is_logical_drive(s):
        return s
    raise ArgumentTypeError("X:\\")


def is_physical_drive(s: str) -> bool:
    return physical_drive_p.match(s) is not None


def is_physical_drive_arg(s: str) -> str:
    if is_physical_drive(s):
        return s
    raise ArgumentTypeError("\\\\?\\PHYSICALDRIVEX where X is a number")


def is_volume_or_drive_arg(s: str) -> str:
    pre = ("\\\\?\\", "\\\\.\\")
    if not s.startswith(pre):
        s = pre[0] + s

    s = s.upper()

    if is_volume_guid_path(s) or is_logical_drive(s) or is_physical_drive(s):
        return s
    else:
        raise ArgumentTypeError("X: or PHYSICALDRIVEX")


def logical_drive_to_win32_file_path(path: str) -> str:
    assert is_logical_drive(path)
    return rf"\\?\{path}"[:-1]


def logical_drive_to_win32_device_path(path: str) -> str:
    assert is_logical_drive(path)
    return rf"\\.\{path}"[:-1]


class EMPTY_BUFFER(Structure):
    pass


def struct_to_array(struct, array):
    tmp = string_at(addressof(struct), sizeof(struct))
    return array(*tmp)


def cstring(buffer: bytes, offset: int = 0, encoding="ascii") -> str:
    end = buffer.find(b"\0", offset)
    return buffer[offset:end].decode(encoding)


def mybyref(value: Optional[Structure]):
    if value is None:
        return None
    return byref(value)


def MyDeviceIoControl(
    DeviceHandle: Any,
    IoControlCode: Any,
    InBuffer: Any,
    OutBuffer: Any,
    *,
    timeout_ms: Optional[int] = None,
    check_output: bool = True,
) -> None:
    BytesReturned = DWORD()

    assert_true("IoControlCode", IoControlCode)

    logger.debug("DeviceIoControl, insize=%d, outsize=%d", sizeof(InBuffer), sizeof(OutBuffer))

    if timeout_ms is None:
        Overlapped = None
    else:
        Overlapped = OVERLAPPED()
        Overlapped.hEvent = CreateEventW(None, True, False, None)

    try:
        try:
            DeviceIoControl(
                DeviceHandle,
                IoControlCode,
                byref(InBuffer),
                sizeof(InBuffer),
                byref(OutBuffer),
                sizeof(OutBuffer),
                byref(BytesReturned),
                mybyref(Overlapped),
            )
        except OSError as e:
            if e.winerror != winerror.ERROR_IO_PENDING:
                raise

            try:
                NumberOfBytesTransferred = DWORD()
                bAlertable = True
                GetOverlappedResultEx(
                    DeviceHandle, byref(Overlapped), byref(NumberOfBytesTransferred), timeout_ms, bAlertable
                )
            except Exception:
                CancelIo(DeviceHandle)  # cancel I/O but don't wait to see if it's actually canceled
                raise

        if check_output and BytesReturned.value != sizeof(OutBuffer):
            raise RuntimeError(
                f"DeviceIoControl expected {sizeof(OutBuffer)} bytes but got {BytesReturned.value} bytes"
            )
    finally:
        if timeout_ms is not None:
            CloseHandle(Overlapped.hEvent)


def open_device(FileName, mode: str = "r", overlapped: bool = False) -> HANDLE:
    DesiredAccess = _mode2access[mode]

    if overlapped:
        FlagsAndAttributes = FILE_FLAG_OVERLAPPED
    else:
        FlagsAndAttributes = 0

    return CreateFileW(
        FileName, DesiredAccess, FILE_SHARE_READ | FILE_SHARE_WRITE, None, OPEN_EXISTING, FlagsAndAttributes, None
    )


def partition_info_to_json(p: winioctl.PARTITION_INFORMATION_EX) -> dict:
    if p.PartitionStyle == winioctl.PARTITION_STYLE.PARTITION_STYLE_MBR:  # MBR
        return {
            "PartitionStyle": p.PartitionStyle.to_enum(),
            "StartingOffset": p.StartingOffset,
            "PartitionLength": p.PartitionLength,
            "PartitionNumber": p.PartitionNumber,
            "RewritePartition": bool(p.RewritePartition.value),
            "IsServicePartition": bool(p.IsServicePartition.value),
            "PartitionType": PartitionInt(p.Mbr.PartitionType),
            "BootIndicator": bool(p.Mbr.BootIndicator.value),
            "RecognizedPartition": bool(p.Mbr.RecognizedPartition),
            "HiddenSectors": p.Mbr.HiddenSectors,
            "PartitionId": str(p.Mbr.PartitionId),
        }
    elif p.PartitionStyle == winioctl.PARTITION_STYLE.PARTITION_STYLE_GPT:  # GPT
        return {
            "PartitionStyle": p.PartitionStyle.to_enum(),
            "StartingOffset": p.StartingOffset,
            "PartitionLength": p.PartitionLength,
            "PartitionNumber": p.PartitionNumber,
            "RewritePartition": bool(p.RewritePartition.value),
            "IsServicePartition": bool(p.IsServicePartition.value),
            "PartitionType": PartitionGuid(p.Gpt.PartitionType),
            "PartitionId": str(p.Gpt.PartitionId),
            "Attributes": p.Gpt.Attributes,
            "Name": p.Gpt.Name,
        }
    else:
        return {}


def drive_layout_to_json(OutBuffer: winioctl.DRIVE_LAYOUT_INFORMATION_EX) -> dict:
    p_style = winioctl.PARTITION_STYLE(OutBuffer.PartitionStyle).to_enum()
    p_info = {"PartitionStyle": p_style}

    if p_style == winioctl.PARTITION_STYLE.PARTITION_STYLE_MBR:  # MBR
        p_info.update(struct2dict(OutBuffer.Mbr))
        p_info["PartitionEntry"] = []

        for p in OutBuffer.PartitionEntry:
            if p.Mbr.PartitionType == winioctl.PARTITION_ENTRY_UNUSED:
                continue

            p_info["PartitionEntry"].append(partition_info_to_json(p))
    elif p_style == winioctl.PARTITION_STYLE.PARTITION_STYLE_GPT:  # GPT
        p_info.update(struct2dict(OutBuffer.Gpt))
        p_info["PartitionEntry"] = []

        for p in OutBuffer.PartitionEntry:
            if p.Gpt.PartitionType == diskguid.PARTITION_ENTRY_UNUSED_GUID:
                continue

            p_info["PartitionEntry"].append(partition_info_to_json(p))

    return p_info


CREATE_FILE_SUPPORTED_PATHS = r"""Volumes:
  Logical drive
- \\?\C:
  Volume GUID path
- \\?\Volume{xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx}
  By disk and volume/partition id
- \\?\HarddiskVolume1
- \\?\CdRom0
- \\?\Harddisk0Partition1
- \\?\GLOBALROOT\Device\HarddiskVolume1
- \\?\GLOBALROOT\Device\Harddisk0\Partition1

Drives:
  By physical drive id
- \\?\PhysicalDrive0
  By physical drive id and DR#
- \\?\GLOBALROOT\Device\Harddisk0\DR0
  Physical device object name:
- \\?\GLOBALROOT\Device\00000001
  Device (interface) paths:
- \\?\usbstor#disk&...
- \\?\SCSI#Disk&...
- \\?\STORAGE#Volume#...
"""


def format_doc_string(**sub):
    def decorator(func):
        func.__doc__ = func.__doc__.format(**sub)
        return func

    return decorator


class StorageMixin:
    @classmethod
    @format_doc_string(CREATE_FILE_SUPPORTED_PATHS=indent(CREATE_FILE_SUPPORTED_PATHS, "        "))
    def from_raw_path(cls, path: str, mode: str = "r", timeout_ms: Optional[int] = None) -> Self:
        """
                Open using paths like

        {CREATE_FILE_SUPPORTED_PATHS}

                or any other path valid for `CreateFile`
        """

        overlapped = timeout_ms is not None
        return cls(open_device(path, mode, overlapped), timeout_ms=timeout_ms)

    # IOCTL_DISK

    def partition_info(self, asjson=True) -> Union[dict, winioctl.PARTITION_INFORMATION_EX]:
        """Returns for:
        - volumes: the information of the partition the volume is on
        - drives: the information for partition 0
            (which is not a valid partition, but has the same length information as the full disk)

        Same length value as from `length_info`.
        """

        InBuffer = EMPTY_BUFFER()
        OutBuffer = winioctl.PARTITION_INFORMATION_EX()
        MyDeviceIoControl(
            self.handle, winioctl.IOCTL_DISK_GET_PARTITION_INFO_EX, InBuffer, OutBuffer, timeout_ms=self.timeout_ms
        )
        if asjson:
            return partition_info_to_json(OutBuffer)
        else:
            return OutBuffer

    def drive_layout(self, max_partitions=128, asjson=True) -> Union[dict, winioctl.DRIVE_LAYOUT_INFORMATION_EX]:
        """Accepts drive and volume handles, but always returns related drive info."""

        InBuffer = EMPTY_BUFFER()
        OutBuffer = winioctl.DRIVE_LAYOUT_INFORMATION_EX_SIZE(max_partitions)()
        MyDeviceIoControl(self.handle, winioctl.IOCTL_DISK_GET_DRIVE_LAYOUT_EX, InBuffer, OutBuffer, check_output=False)
        if asjson:
            return drive_layout_to_json(OutBuffer)
        else:
            return OutBuffer

    def length_info(self) -> int:
        """Returns for:
        - volumes: the full partition size
            (might be larger than than the filesystem size ie. disk_usage().total)
        - drives: the full drive size
            (drive size might be different if computed from drive geometry,
            but the geometry is often incorrect)

        Requires read access rights.
        Same as length value in `partition_info`.
        """

        InBuffer = EMPTY_BUFFER()
        OutBuffer = winioctl.GET_LENGTH_INFORMATION()
        MyDeviceIoControl(
            self.handle, winioctl.IOCTL_DISK_GET_LENGTH_INFO, InBuffer, OutBuffer, timeout_ms=self.timeout_ms
        )
        return OutBuffer.Length

    def drive_geometry(self) -> dict:
        """Accepts drive and volume handles, but always returns related drive info."""

        InBuffer = EMPTY_BUFFER()
        OutBuffer = winioctl.DISK_GEOMETRY()
        MyDeviceIoControl(
            self.handle, winioctl.IOCTL_DISK_GET_DRIVE_GEOMETRY, InBuffer, OutBuffer, timeout_ms=self.timeout_ms
        )
        return struct2dict(OutBuffer)

    # IOCTL_STORAGE

    def storage_query_property(self, property_id: winioctl.STORAGE_PROPERTY_ID, asjson=True) -> Union[dict, Structure]:
        mapping = {
            winioctl.STORAGE_PROPERTY_ID.StorageAdapterProperty: winioctl.STORAGE_ADAPTER_DESCRIPTOR,
            winioctl.STORAGE_PROPERTY_ID.StorageAccessAlignmentProperty: winioctl.STORAGE_ACCESS_ALIGNMENT_DESCRIPTOR,
            winioctl.STORAGE_PROPERTY_ID.StorageDeviceSeekPenaltyProperty: winioctl.DEVICE_SEEK_PENALTY_DESCRIPTOR,
            winioctl.STORAGE_PROPERTY_ID.StorageDeviceTrimProperty: winioctl.DEVICE_TRIM_DESCRIPTOR,
            winioctl.STORAGE_PROPERTY_ID.StorageDeviceTemperatureProperty: winioctl.STORAGE_TEMPERATURE_DATA_DESCRIPTOR,
        }
        Descriptor = mapping[property_id]

        InBuffer = winioctl.STORAGE_PROPERTY_QUERY()
        OutBuffer = Descriptor()
        InBuffer.PropertyId = property_id
        InBuffer.QueryType = winioctl.STORAGE_QUERY_TYPE.PropertyStandardQuery
        MyDeviceIoControl(
            self.handle, winioctl.IOCTL_STORAGE_QUERY_PROPERTY, InBuffer, OutBuffer, timeout_ms=self.timeout_ms
        )
        if asjson:
            return struct2dict(OutBuffer)
        else:
            return OutBuffer

    def sqp_adapter(self) -> dict:
        """Accepts drive and volume handles, but always returns related drive info."""

        out = self.storage_query_property(winioctl.STORAGE_PROPERTY_ID.StorageAdapterProperty)
        out["BusType"] = ntddstor.STORAGE_BUS_TYPE(out["BusType"]).to_enum()
        out["SrbType"] = SrbType(out["SrbType"])
        out["AddressType"] = StorageAddressType(out["AddressType"])
        return out

    def sqp_alignment(self) -> dict:
        """Accepts drive and volume handles, but always returns related drive info."""

        return self.storage_query_property(winioctl.STORAGE_PROPERTY_ID.StorageAccessAlignmentProperty)

    def sqp_seek_penalty(self) -> bool:
        """Accepts drive and volume handles, but always returns related drive info."""

        return bool(
            self.storage_query_property(
                winioctl.STORAGE_PROPERTY_ID.StorageDeviceSeekPenaltyProperty, False
            ).IncursSeekPenalty
        )

    def sqp_trim_enabled(self) -> bool:
        """Accepts drive and volume handles, but always returns related drive info."""

        return bool(
            self.storage_query_property(winioctl.STORAGE_PROPERTY_ID.StorageDeviceTrimProperty, False).TrimEnabled
        )

    def sqp_temperature(self) -> dict:
        """Accepts drive and volume handles, but always returns related drive info."""

        # requires Windows 10+
        return self.storage_query_property(winioctl.STORAGE_PROPERTY_ID.StorageDeviceTemperatureProperty)

    def get_device_number(self) -> dict:
        """Returns for:
        - volumes: device type, device number and partition number
        - drives: device type, device number and partition number 0

        """

        InBuffer = EMPTY_BUFFER()
        OutBuffer = winioctl.STORAGE_DEVICE_NUMBER()
        MyDeviceIoControl(
            self.handle, winioctl.IOCTL_STORAGE_GET_DEVICE_NUMBER, InBuffer, OutBuffer, timeout_ms=self.timeout_ms
        )
        out = struct2dict(OutBuffer)
        out["DeviceType"] = DEVICE_TYPE(out["DeviceType"])
        return out

    def read_capacity(self) -> dict:
        """Accepts drive and volume handles, but always returns related drive info.
        The DiskLength returned is the same as if `length_info` is used on the related drive.

        Requires read access rights.
        """

        InBuffer = EMPTY_BUFFER()
        OutBuffer = winioctl.STORAGE_READ_CAPACITY()
        MyDeviceIoControl(
            self.handle, winioctl.IOCTL_STORAGE_READ_CAPACITY, InBuffer, OutBuffer, timeout_ms=self.timeout_ms
        )
        return struct2dict(OutBuffer)

    def sqp_device(self) -> dict:
        """Accepts drive and volume handles, but always returns related drive info."""

        # obtain buffer size first
        InBuffer = winioctl.STORAGE_PROPERTY_QUERY()
        OutBuffer = winioctl.STORAGE_DESCRIPTOR_HEADER()
        InBuffer.PropertyId = winioctl.STORAGE_PROPERTY_ID.StorageDeviceProperty
        InBuffer.QueryType = winioctl.STORAGE_QUERY_TYPE.PropertyStandardQuery
        MyDeviceIoControl(
            self.handle, winioctl.IOCTL_STORAGE_QUERY_PROPERTY, InBuffer, OutBuffer, timeout_ms=self.timeout_ms
        )

        # query with correct buffer size
        InBuffer = winioctl.STORAGE_PROPERTY_QUERY()
        OutBuffer = winioctl.STORAGE_DEVICE_DESCRIPTOR_SIZE(
            OutBuffer.Size - sizeof(winioctl.STORAGE_DEVICE_DESCRIPTOR)
        )()
        InBuffer.PropertyId = winioctl.STORAGE_PROPERTY_ID.StorageDeviceProperty
        InBuffer.QueryType = winioctl.STORAGE_QUERY_TYPE.PropertyStandardQuery

        MyDeviceIoControl(
            self.handle, winioctl.IOCTL_STORAGE_QUERY_PROPERTY, InBuffer, OutBuffer, timeout_ms=self.timeout_ms
        )

        out = {
            "DeviceType": ScsiDriveType(OutBuffer.DeviceType),
            "DeviceTypeModifier": OutBuffer.DeviceTypeModifier,
            "RemovableMedia": bool(OutBuffer.RemovableMedia.value),
            "CommandQueueing": bool(OutBuffer.CommandQueueing.value),
            "BusType": OutBuffer.BusType.to_enum(),
        }

        raw = bytes(OutBuffer.RawDeviceProperties)
        offset = sizeof(winioctl.STORAGE_DEVICE_DESCRIPTOR_SIZE(0)())
        assert offset == 36
        assert OutBuffer.RawPropertiesLength in (0, offset)

        if OutBuffer.VendorIdOffset == 0:
            out["VendorId"] = None
        else:
            out["VendorId"] = cstring(raw, OutBuffer.VendorIdOffset - offset)

        if OutBuffer.ProductIdOffset == 0:
            out["ProductId"] = None
        else:
            out["ProductId"] = cstring(raw, OutBuffer.ProductIdOffset - offset)

        if OutBuffer.ProductRevisionOffset == 0:
            out["ProductRevision"] = None
        else:
            out["ProductRevision"] = cstring(raw, OutBuffer.ProductRevisionOffset - offset)

        if OutBuffer.SerialNumberOffset == 0:
            out["SerialNumber"] = None
        else:
            out["SerialNumber"] = cstring(raw, OutBuffer.SerialNumberOffset - offset)

        return out


class Storage(WindowsHandle, StorageMixin):
    pass


class Volume(WindowsHandle, StorageMixin):
    @classmethod
    def from_win32_drive_letter_p(cls, path: str, mode: str = "r") -> Self:
        assert win32_drive_letter_p.match(path)
        return cls(open_device(path, mode))

    @classmethod
    def from_logical_drive(cls, path: str, mode: str = "r") -> Self:
        return cls(open_device(logical_drive_to_win32_device_path(path), mode))

    @classmethod
    def from_volume_guid_path(cls, path: str, mode: str = "r") -> Self:
        if not is_volume_guid_path(path):
            raise ValueError("path is not a volume guid path")

        volume_guid_path = path[:-1]  # important: remove trailing slash
        return cls(open_device(volume_guid_path, mode))

    @classmethod
    def from_harddisk_volume_index(cls, index: int, mode: str = "r") -> Self:
        return cls(open_device(f"\\\\.\\HarddiskVolume{index}", mode))

    @classmethod
    def from_disk_index_and_partition(cls, index: int, partition: int, mode: str = "r") -> Self:
        return cls(open_device(f"\\\\?\\Harddisk{index}Partition{partition}", mode))

    # FSCTL

    def lock(self) -> None:
        MyDeviceIoControl(self.handle, winioctl.FSCTL_LOCK_VOLUME, EMPTY_BUFFER(), EMPTY_BUFFER())

    def unlock(self) -> None:
        MyDeviceIoControl(self.handle, winioctl.FSCTL_UNLOCK_VOLUME, EMPTY_BUFFER(), EMPTY_BUFFER())

    def dismount(self) -> None:
        MyDeviceIoControl(self.handle, winioctl.FSCTL_DISMOUNT_VOLUME, EMPTY_BUFFER(), EMPTY_BUFFER())

    def is_mounted(self) -> bool:
        try:
            MyDeviceIoControl(self.handle, winioctl.FSCTL_IS_VOLUME_MOUNTED, EMPTY_BUFFER(), EMPTY_BUFFER())
            return True
        except OSError as e:
            if e.winerror == winerror.ERROR_NOT_READY:
                return False
            else:
                raise

    def extend(self) -> None:
        """Allows to read beyond the filesytem volume to read the full partition size,
        which might be slightly larger.
        Note: Might require unbuffered IO and manual management of read size,
            since the Python read() buffering can do out of bounds reads.
            Get the correct partition size with `length_info` or `partition_info`.
        """
        MyDeviceIoControl(self.handle, winioctl.FSCTL_ALLOW_EXTENDED_DASD_IO, EMPTY_BUFFER(), EMPTY_BUFFER())

    def filesystem_statistics(self) -> List[dict]:
        """Retrieves the information from various file system performance counters."""

        class FILESYSTEM_STATISTICS_NTFS(Structure):
            _fields_ = [
                ("fss", winioctl.FILESYSTEM_STATISTICS),
                ("fs", winioctl.NTFS_STATISTICS),
                ("padding", BYTE * 52),
            ]

        class FILESYSTEM_STATISTICS_FAT(Structure):
            _fields_ = [
                ("fss", winioctl.FILESYSTEM_STATISTICS),
                ("fs", winioctl.FAT_STATISTICS),
                ("padding", BYTE * 36),
            ]

        class FILESYSTEM_STATISTICS_EXFAT(Structure):
            _fields_ = [
                ("fss", winioctl.FILESYSTEM_STATISTICS),
                ("fs", winioctl.EXFAT_STATISTICS),
                ("padding", BYTE * 36),
            ]

        assert sizeof(FILESYSTEM_STATISTICS_NTFS) % 64 == 0
        assert sizeof(FILESYSTEM_STATISTICS_FAT) % 64 == 0
        assert sizeof(FILESYSTEM_STATISTICS_EXFAT) % 64 == 0

        InBuffer = EMPTY_BUFFER()
        OutBuffer = winioctl.FILESYSTEM_STATISTICS()
        try:
            MyDeviceIoControl(
                self.handle, winioctl.FSCTL_FILESYSTEM_GET_STATISTICS, InBuffer, OutBuffer, timeout_ms=self.timeout_ms
            )
        except OSError as e:
            if e.winerror != winerror.ERROR_MORE_DATA:
                raise

            fsmap = {
                FILESYSTEM_STATISTICS_TYPE.NTFS.value: FILESYSTEM_STATISTICS_NTFS,
                FILESYSTEM_STATISTICS_TYPE.FAT.value: FILESYSTEM_STATISTICS_FAT,
                FILESYSTEM_STATISTICS_TYPE.EXFAT.value: FILESYSTEM_STATISTICS_EXFAT,
            }
            OutBuffer = (fsmap[OutBuffer.FileSystemType] * os.cpu_count())()
            MyDeviceIoControl(
                self.handle, winioctl.FSCTL_FILESYSTEM_GET_STATISTICS, InBuffer, OutBuffer, timeout_ms=self.timeout_ms
            )

        out = [struct2dict(s) for s in OutBuffer]
        for d in out:
            d.pop("padding")
        return out

    def ntfs_volume_data(self) -> dict:
        class NTFS_VOLUME_DATA_COMBINED(Structure):
            _fields_ = [
                ("basic", winioctl.NTFS_VOLUME_DATA_BUFFER),
                ("extended", winioctl.NTFS_EXTENDED_VOLUME_DATA),
            ]

        InBuffer = EMPTY_BUFFER()
        OutBuffer = NTFS_VOLUME_DATA_COMBINED()
        MyDeviceIoControl(
            self.handle, winioctl.FSCTL_GET_NTFS_VOLUME_DATA, InBuffer, OutBuffer, timeout_ms=self.timeout_ms
        )
        return struct2dict(OutBuffer)

    def is_dirty(self):
        InBuffer = EMPTY_BUFFER()
        OutBuffer = ULONG()
        MyDeviceIoControl(self.handle, winioctl.FSCTL_IS_VOLUME_DIRTY, InBuffer, OutBuffer, timeout_ms=self.timeout_ms)
        return IsDirtyFlags(OutBuffer.value)

    # IOCTL_VOLUME

    def bring_online(self) -> None:
        MyDeviceIoControl(self.handle, winioctl.IOCTL_VOLUME_ONLINE, EMPTY_BUFFER(), EMPTY_BUFFER())

    def take_offline(self) -> None:
        MyDeviceIoControl(self.handle, winioctl.IOCTL_VOLUME_OFFLINE, EMPTY_BUFFER(), EMPTY_BUFFER())


class Drive(WindowsHandle, StorageMixin):
    @classmethod
    def from_drive_type_and_index(cls, drive_type: str, drive_index: int, mode: str = "r") -> Self:
        if drive_type.lower() not in ("physicaldrive", "scsi"):
            msg = f"Invalid drive type: {drive_type}"
            raise ValueError(msg)

        return cls(open_device(rf"\\.\{drive_type}{drive_index}", mode))

    @classmethod
    def from_scsi_index(cls, drive_index: int, mode: str = "r") -> Self:
        return cls(open_device(rf"\\.\Scsi{drive_index}:", mode))

    # IOCTL_DISK

    def verify_extent(self, start: int, length: int) -> None:
        """Can cause issues with some drivers (like TrueCrypt)."""

        InBuffer = winioctl.VERIFY_INFORMATION()
        OutBuffer = EMPTY_BUFFER()

        InBuffer.StartingOffset = start
        InBuffer.Length = length

        MyDeviceIoControl(self.handle, winioctl.IOCTL_DISK_VERIFY, InBuffer, OutBuffer, timeout_ms=self.timeout_ms)

    def is_writable(self) -> bool:
        try:
            MyDeviceIoControl(self.handle, winioctl.IOCTL_DISK_IS_WRITABLE, EMPTY_BUFFER(), EMPTY_BUFFER())
            return True
        except OSError as e:
            if e.winerror == winerror.ERROR_WRITE_PROTECT:
                return False
            else:
                raise

    def cache_information(self):
        InBuffer = EMPTY_BUFFER()
        OutBuffer = winioctl.DISK_CACHE_INFORMATION()

        MyDeviceIoControl(
            self.handle, winioctl.IOCTL_DISK_GET_CACHE_INFORMATION, InBuffer, OutBuffer, timeout_ms=self.timeout_ms
        )

        return struct2dict(OutBuffer)

    # IOCTL_STORAGE

    def sqp_device_protocol_specific(
        self,
        protocol_type: winioctl.STORAGE_PROTOCOL_TYPE,
        data_type: Union[winioctl.STORAGE_PROTOCOL_ATA_DATA_TYPE, winioctl.STORAGE_PROTOCOL_NVME_DATA_TYPE],
        request_value: Union[nvme.NVME_IDENTIFY_CNS_CODES, nvme.NVME_LOG_PAGES],
    ) -> dict:
        psd_size = sizeof(winioctl.STORAGE_PROTOCOL_SPECIFIC_DATA)

        if protocol_type == winioctl.STORAGE_PROTOCOL_TYPE.ProtocolTypeNvme:
            if data_type == winioctl.STORAGE_PROTOCOL_NVME_DATA_TYPE.NVMeDataTypeIdentify:
                mapping = {
                    nvme.NVME_IDENTIFY_CNS_CODES.NVME_IDENTIFY_CNS_SPECIFIC_NAMESPACE: nvme.NVME_IDENTIFY_NAMESPACE_DATA,
                    nvme.NVME_IDENTIFY_CNS_CODES.NVME_IDENTIFY_CNS_CONTROLLER: nvme.NVME_IDENTIFY_CONTROLLER_DATA,
                }

            elif data_type == winioctl.STORAGE_PROTOCOL_NVME_DATA_TYPE.NVMeDataTypeLogPage:
                mapping = {
                    nvme.NVME_LOG_PAGES.NVME_LOG_PAGE_HEALTH_INFO: nvme.NVME_HEALTH_INFO_LOG,
                }
            else:
                raise ValueError(f"Unsupported data_type: {data_type}")
        else:
            raise ValueError(f"Unsupported protocol_type: {protocol_type}")

        try:
            ProtocolData = mapping[request_value]
        except KeyError:
            raise ValueError(f"Unsupported request_value: {request_value}")

        class DUMMY(Structure):
            _fields_ = [("spq", winioctl.STORAGE_PROPERTY_QUERY_SIZE(psd_size)), ("pd", ProtocolData)]

        InBuffer = DUMMY()

        psd = winioctl.STORAGE_PROTOCOL_SPECIFIC_DATA()
        psd.ProtocolType = protocol_type
        psd.DataType = data_type
        psd.ProtocolDataRequestValue = request_value  # cdw10
        # psd.ProtocolDataRequestSubValue = 1  # If the ProtocolDataRequestValue is NVME_IDENTIFY_CNS_SPECIFIC_NAMESPACE, this will have a value of the namespace ID.
        psd.ProtocolDataOffset = psd_size  # DUMMY.pd.offset gives error 87
        psd.ProtocolDataLength = sizeof(ProtocolData)

        # InBuffer.PropertyId = winioctl.STORAGE_PROPERTY_ID.StorageAdapterProtocolSpecificProperty
        InBuffer.spq.PropertyId = winioctl.STORAGE_PROPERTY_ID.StorageDeviceProtocolSpecificProperty
        InBuffer.spq.QueryType = winioctl.STORAGE_QUERY_TYPE.PropertyStandardQuery
        InBuffer.spq.AdditionalParameters = struct_to_array(psd, (BYTE * psd_size))
        assert sizeof(InBuffer.spq) == 48

        MyDeviceIoControl(self.handle, winioctl.IOCTL_STORAGE_QUERY_PROPERTY, InBuffer, InBuffer)

        out = {
            "PropertyId": InBuffer.spq.PropertyId.value,
            "QueryType": InBuffer.spq.QueryType.value,
            "AdditionalParameters": struct2dict(psd),
            ProtocolData.__name__: struct2dict(InBuffer.pd),
        }

        out["AdditionalParameters"]["DataType"] = winioctl.STORAGE_PROTOCOL_NVME_DATA_TYPE(
            out["AdditionalParameters"]["DataType"]
        ).to_enum()
        if data_type == winioctl.STORAGE_PROTOCOL_NVME_DATA_TYPE.NVMeDataTypeIdentify:
            out["AdditionalParameters"]["ProtocolDataRequestValue"] = nvme.NVME_IDENTIFY_CNS_CODES(
                out["AdditionalParameters"]["ProtocolDataRequestValue"]
            ).to_enum()
        elif data_type == winioctl.STORAGE_PROTOCOL_NVME_DATA_TYPE.NVMeDataTypeLogPage:
            out["AdditionalParameters"]["ProtocolDataRequestValue"] = nvme.NVME_LOG_PAGES(
                out["AdditionalParameters"]["ProtocolDataRequestValue"]
            ).to_enum()

        assert psd.FixedProtocolReturnData == 0, psd.FixedProtocolReturnData
        return out

    def firmware_get_info(self):
        """Query detailed firmware information."""

        InBuffer = winioctl.STORAGE_HW_FIRMWARE_INFO_QUERY()
        OutBuffer = winioctl.STORAGE_HW_FIRMWARE_INFO()

        InBuffer.Size = sizeof(winioctl.STORAGE_HW_FIRMWARE_INFO_QUERY)

        MyDeviceIoControl(
            self.handle, winioctl.IOCTL_STORAGE_FIRMWARE_GET_INFO, InBuffer, OutBuffer, timeout_ms=self.timeout_ms
        )

        if OutBuffer.SlotCount > 1:
            logger.error("Device has more than one firmware slot. We only return the first one.")

        return struct2dict(OutBuffer)

    def predict_failure(self):
        InBuffer = EMPTY_BUFFER()
        OutBuffer = ntddstor.STORAGE_PREDICT_FAILURE()

        MyDeviceIoControl(
            self.handle, winioctl.IOCTL_STORAGE_PREDICT_FAILURE, InBuffer, OutBuffer, timeout_ms=self.timeout_ms
        )

        return struct2dict(OutBuffer)

    # IOCTL misc

    def get_smart_version(self) -> dict:
        """Usually only supported by ATA devices, ie. not by NVME or USB attached drives."""

        InBuffer = EMPTY_BUFFER()
        OutBuffer = winioctl.GETVERSIONINPARAMS()

        MyDeviceIoControl(self.handle, winioctl.SMART_GET_VERSION, InBuffer, OutBuffer, timeout_ms=self.timeout_ms)

        return struct2dict(OutBuffer)

    def get_scsi_address(self):
        InBuffer = EMPTY_BUFFER()
        OutBuffer = ntddscsi.SCSI_ADDRESS()

        MyDeviceIoControl(self.handle, ntddscsi.IOCTL_SCSI_GET_ADDRESS, InBuffer, OutBuffer, timeout_ms=self.timeout_ms)

        return struct2dict(OutBuffer)


def query_dos_devices(path: Optional[str] = None) -> List[str]:
    """Acceptable input paths:
    - `C:` -> '\\Device\\HarddiskVolume1'
    - `Volume{xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx}` -> '\\Device\\HarddiskVolume1'
    - `HarddiskVolume1` -> `\\Device\\HarddiskVolume1`
    - `Harddisk0Partition1` -> `\\Device\\HarddiskVolume1`
    - `CdRom0` -> `\\Device\\CdRom0`
    - `GLOBALROOT\\Device\\Harddisk0\\Partition1` -> `\\Device\\HarddiskVolume1`
    - `PhysicalDrive0` -> `\\Device\\Harddisk0\\DR0`
    - `scsi#disk...` -> `\\Device\\00000001`
    Not supported:
    - `GLOBALROOT\\Device\\HarddiskVolume1`
    - `GLOBALROOT\\Device\\Harddisk0\\DR0`
    - `GLOBALROOT\\Device\\00000001`
    """

    if path is not None:
        if path.endswith("\\"):
            raise ValueError("Trailing \\ are not supported")
        if path.startswith("\\\\?\\"):
            raise ValueError("Leading \\\\?\\ is not supported")

    ucchMax = 64 * 1024
    TargetPath = create_unicode_buffer(ucchMax)
    BufferLength = QueryDosDeviceW(path, TargetPath, ucchMax)

    result = memoryview(TargetPath)[: BufferLength - 2].tobytes().decode("utf-16-le")
    return result.split("\0")


def find_volumes() -> Iterator[str]:
    """Yields volume guid paths like `\\\\?\\Volume{xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx}\\`
    `os.listvolumes()` on Python 3.12+
    """

    BufferLength = 64 * 1024
    VolumeName = create_unicode_buffer(BufferLength)

    try:
        handle = FindFirstVolumeW(VolumeName, BufferLength)
    except OSError as e:
        if e.winerror == winerror.ERROR_NO_MORE_FILES:
            return
        raise

    try:
        while True:
            try:
                FindNextVolumeW(handle, VolumeName, BufferLength)
                yield VolumeName.value
            except OSError as e:
                if e.winerror == winerror.ERROR_NO_MORE_FILES:
                    break
                raise
    finally:
        FindVolumeClose(handle)


def find_volume_mount_points(volume_guid_path: str) -> List[str]:
    if not is_volume_guid_path(volume_guid_path):
        msg = f"Not a volume guid path: {volume_guid_path}"
        raise ValueError(msg)

    BufferLength = 64 * 1024
    VolumeMountPoint = create_unicode_buffer(BufferLength)

    try:
        handle = FindFirstVolumeMountPointW(volume_guid_path, VolumeMountPoint, BufferLength)
    except OSError as e:
        if e.winerror == winerror.ERROR_NO_MORE_FILES:
            return
        raise

    try:
        while True:
            try:
                FindNextVolumeMountPointW(handle, VolumeMountPoint, BufferLength)
                yield VolumeMountPoint.value
            except OSError as e:
                if e.winerror == winerror.ERROR_NO_MORE_FILES:
                    break
                raise
    finally:
        FindVolumeMountPointClose(handle)


def get_logical_drives() -> List[str]:
    """Get logical drives. Eg. ['C:\\\\', 'D:\\\\']
    Same as `os.listdrives()` on Python 3.12+
    """

    # BufferLength = GetLogicalDriveStringsW(0, None)
    BufferLength = 32 * 1024 - 1  # GetLogicalDriveStringsW fails with buffers >= 32k
    Buffer = create_unicode_buffer(BufferLength)
    ReturnLength = GetLogicalDriveStringsW(BufferLength, Buffer)
    # ReturnLength doesn't include the terminating null character
    assert ReturnLength < BufferLength
    result = memoryview(Buffer)[:ReturnLength].tobytes().decode("utf-16-le")
    if result:
        return result[:-1].split("\0")
    else:
        return []


def get_volume_path_names(volume_guid_path: str) -> List[str]:
    """Retrieves a list of drive letters and mounted folder paths for the specified volume.
    Same as `os.listmounts()` on Python 3.12+
    """

    if not is_volume_guid_path(volume_guid_path):
        msg = f"Not a volume guid path: {volume_guid_path}"
        raise ValueError(msg)

    BufferLength = 64 * 1024
    VolumePathNames = create_unicode_buffer(BufferLength)
    ReturnLength = DWORD()  # full buffer size including terminating null character

    GetVolumePathNamesForVolumeNameW(volume_guid_path, VolumePathNames, BufferLength, byref(ReturnLength))
    result = memoryview(VolumePathNames)[: ReturnLength.value - 1].tobytes().decode("utf-16-le")
    if result:
        return result[:-1].split("\0")
    else:
        return []


def get_volume_name(VolumeMountPoint: str) -> str:
    """Retrieves a volume GUID path for the volume that is associated with the specified volume mount point ( drive letter, volume GUID path, or mounted folder)."""

    BufferLength = 64 * 1024
    VolumeName = create_unicode_buffer(BufferLength)
    GetVolumeNameForVolumeMountPointW(VolumeMountPoint, VolumeName, BufferLength)
    return VolumeName.value


def enum_device_paths(
    setup_class: Optional[GUID] = None,
    interface_class: Optional[GUID] = None,
    enumerator: Optional[str] = None,
    device_instance_id: Optional[str] = None,
    default: bool = False,
    present: bool = True,
    profile: bool = False,
) -> Tuple[List[dict], List[str]]:
    flags = 0
    if default:
        flags |= setupapi.DIGCF_DEFAULT
    if present:
        flags |= setupapi.DIGCF_PRESENT
    if profile:
        flags |= setupapi.DIGCF_PROFILE

    if setup_class is None and interface_class is None:
        if enumerator is not None:
            enumerator = create_unicode_buffer(enumerator)

        hDevInfo = setupapi.SetupDiGetClassDevsW(None, enumerator, None, flags | setupapi.DIGCF_ALLCLASSES)
    elif setup_class is not None and interface_class is None:
        if enumerator is not None:
            enumerator = create_unicode_buffer(enumerator)

        hDevInfo = setupapi.SetupDiGetClassDevsW(byref(setup_class), enumerator, None, flags)
    elif setup_class is None and interface_class is not None:
        if device_instance_id is not None:
            device_instance_id = create_unicode_buffer(device_instance_id)

        hDevInfo = setupapi.SetupDiGetClassDevsW(
            byref(interface_class), device_instance_id, None, flags | setupapi.DIGCF_DEVICEINTERFACE
        )
    else:
        raise ValueError()

    device_infos = []
    device_paths = []

    try:
        if interface_class is None:
            DeviceInfoData = setupapi.SP_DEVINFO_DATA()
            DeviceInfoData.cbSize = sizeof(setupapi.SP_DEVINFO_DATA)

            while True:
                try:
                    setupapi.SetupDiEnumDeviceInfo(hDevInfo, len(device_infos), byref(DeviceInfoData))
                except OSError as e:
                    if e.winerror == winerror.ERROR_NO_MORE_ITEMS:
                        break
                    raise

                ClassDescription = create_unicode_buffer(setupapi.LINE_LEN)
                try:
                    setupapi.SetupDiGetClassDescriptionW(
                        byref(DeviceInfoData.ClassGuid), ClassDescription, setupapi.LINE_LEN, None
                    )
                except OSError as e:
                    if e.winerror == -536870394:  # ERROR_INVALID_CLASS
                        class_description = None
                    else:
                        raise
                else:
                    class_description = ClassDescription.value

                PropertyType = devpropdef.DEVPROPTYPE()
                PropertyBuffer = (BYTE * 4)()
                PropertyBufferSize = 4
                RequiredSize = DWORD()

                try:
                    setupapi.SetupDiGetDevicePropertyW(
                        hDevInfo,
                        byref(DeviceInfoData),
                        devpkey.DEVPKEY_Device_ProblemCode,
                        byref(PropertyType),
                        cast(PropertyBuffer, PBYTE),
                        PropertyBufferSize,
                        byref(RequiredSize),
                        0,
                    )
                    assert PropertyType.value == devpropdef.DEVPROP_TYPE_UINT32
                    devprop = cast(PropertyBuffer, POINTER(UINT32))[0]
                except OSError as e:
                    if e.winerror == winerror.ERROR_NOT_FOUND:
                        devprop = None
                    else:
                        raise

                device_infos.append(
                    {
                        "class-guid": str(DeviceInfoData.ClassGuid),
                        "device-instance": DeviceInfoData.DevInst,
                        "class-description": class_description,
                        "problem-code": devprop,
                    }
                )
        else:
            DeviceInterfaceData = setupapi.SP_DEVICE_INTERFACE_DATA()
            DeviceInterfaceData.cbSize = sizeof(setupapi.SP_DEVICE_INTERFACE_DATA)
            requiredSize = DWORD()

            while True:
                try:
                    setupapi.SetupDiEnumDeviceInterfaces(
                        hDevInfo, None, byref(interface_class), len(device_paths), byref(DeviceInterfaceData)
                    )
                except OSError as e:
                    if e.winerror == winerror.ERROR_NO_MORE_ITEMS:
                        break
                    raise
                assert DeviceInterfaceData.InterfaceClassGuid == interface_class

                try:
                    setupapi.SetupDiGetDeviceInterfaceDetailW(
                        hDevInfo, byref(DeviceInterfaceData), None, 0, byref(requiredSize), None
                    )
                except OSError as e:
                    if e.winerror == winerror.ERROR_INSUFFICIENT_BUFFER:
                        pass
                    else:
                        raise

                size = requiredSize.value - sizeof(setupapi.SP_DEVICE_INTERFACE_DETAIL_DATA_W)
                DeviceInterfaceDetailData = setupapi.SP_DEVICE_INTERFACE_DETAIL_DATA_W_SIZE(size)()
                DeviceInterfaceDetailData.cbSize = sizeof(setupapi.SP_DEVICE_INTERFACE_DETAIL_DATA_W)

                setupapi.SetupDiGetDeviceInterfaceDetailW(
                    hDevInfo,
                    byref(DeviceInterfaceData),
                    cast(byref(DeviceInterfaceDetailData), POINTER(setupapi.SP_DEVICE_INTERFACE_DETAIL_DATA_W)),
                    requiredSize,
                    None,
                    None,
                )

                device_paths.append(DeviceInterfaceDetailData.DevicePath)

    finally:
        setupapi.SetupDiDestroyDeviceInfoList(hDevInfo)

    return device_infos, device_paths


def enum_disks():
    device_infos, device_paths = enum_device_paths(interface_class=winioctl.GUID_DEVINTERFACE_DISK)
    for device_path in device_paths:
        out = {"DevicePath": device_path}
        with Drive.from_raw_path(device_path, "") as drive:
            dn = drive.get_device_number()
            dn.pop("PartitionNumber")
            out.update(dn)
            yield out


def enum_cdrom():
    device_infos, device_paths = enum_device_paths(interface_class=winioctl.GUID_DEVINTERFACE_CDROM)
    for device_path in device_paths:
        out = {"DevicePath": device_path}
        with Drive.from_raw_path(device_path, "") as drive:
            dn = drive.get_device_number()
            dn.pop("PartitionNumber")
            out.update(dn)
            yield out


if __name__ == "__main__":
    from argparse import ArgumentParser
    from pprint import pprint

    parser = ArgumentParser()
    parser.add_argument("driveindex", type=int)
    args = parser.parse_args()

    try:
        drive = Drive.from_drive_type_and_index("PhysicalDrive", args.driveindex, "r")
    except PermissionError:
        logging.warning("Opening drive with read access failed. Using no access.")
        drive = Drive.from_drive_type_and_index("PhysicalDrive", args.driveindex, "")

    with drive as d:
        try:
            print("Alignment:", d.sqp_alignment())
        except OSError as e:
            print("Alignment:", e)
        try:
            print("Seek penalty:", d.sqp_seek_penalty())
        except OSError as e:
            print("Seek penalty:", e)
        try:
            print("Trim enabled:", d.sqp_trim_enabled())
        except OSError as e:
            print("Trim enabled:", e)
        try:
            print("Capacity:", d.read_capacity())
        except OSError as e:
            print("Capacity:", e)
        print("Device:", d.sqp_device())
        try:
            print("Geometry:", d.drive_geometry())
        except OSError as e:
            print("Geometry:", e)
        try:
            print("SMART version:", d.get_smart_version())
        except OSError as e:
            print("SMART version:", e)
        try:
            print("Temperature:", d.sqp_temperature())
        except OSError as e:
            print("Temperature:", e)

        pprint(d.drive_layout())



from typing import Any, Callable

from .func import identity

try:
    from numba import prange as oprange

except ImportError:
    from warnings import warn

    warn("Numba not found.", stacklevel=2)

    oprange = range


def opjit(*args: Any, **kwargs: Any) -> Callable:
    """Optional / opportunistic numba jit decorator."""

    try:
        from numba import njit

    except ImportError:
        from warnings import warn

        warn("Numba not found. Using slower pure Python version.", stacklevel=2)

        return identity

    kwargs.setdefault("fastmath", True)
    kwargs.setdefault("cache", True)

    def dec(func):
        return njit(*args, **kwargs)(func)

    return dec



def force_decode(data: bytes) -> str:
    try:
        return data.decode()  # try default encoding
    except UnicodeDecodeError:
        try:
            return data.decode("utf-8")
        except UnicodeDecodeError:
            return data.decode("latin1")  # should never fail



from typing import Iterator, List

from pptx import Presentation  # python-pptx


def read_ppt(path: str) -> Iterator[List[str]]:
    prs = Presentation(path)
    for slide in prs.slides:
        shapes = [shape.text for shape in slide.shapes if hasattr(shape, "text")]
        yield shapes



import logging
from math import sqrt


# was: FindWidthToHeightRatio
def perspective_rectangle_aspect_ratio(corners, principal_point, focal_length=None):
    """Return the width-to-height ratio of a perspectivly transformed rectangle.
    taken from: http://stackoverflow.com/questions/1194352/proportions-of-a-perspective-deformed-rectangle
    corners: corners of the quadrilateral
    principal point: center of the image (not the quadrilateral)
    """

    def sqr(x):
        return x**2

    # (m1x, m1y), (m2x, m2y), (m3x, m3y), (m4x, m4y) = corners
    (m1x, m1y), (m3x, m3y), (m4x, m4y), (m2x, m2y) = corners
    u0, v0 = principal_point

    # in case it matters: licensed under GPLv2 or later
    # legend:
    # sqr(x)  = x*x
    # sqrt(x) = square root of x

    # let m1x,m1y ... m4x,m4y be the (x,y) pixel coordinates
    # of the 4 corners of the detected quadrangle
    # i.e. (m1x, m1y) are the cordinates of the first corner,
    # (m2x, m2y) of the second corner and so on.
    # let u0, v0 be the pixel coordinates of the principal point of the image
    # for a normal camera this will be the center of the image,
    # i.e. u0=IMAGEWIDTH/2; v0 =IMAGEHEIGHT/2
    # This assumption does not hold if the image has been cropped asymmetrically

    # first, transform the image so the principal point is at (0,0)
    # this makes the following equations much easier
    m1x = m1x - u0
    m1y = m1y - v0
    m2x = m2x - u0
    m2y = m2y - v0
    m3x = m3x - u0
    m3y = m3y - v0
    m4x = m4x - u0
    m4y = m4y - v0

    # temporary variables k2, k3
    k2 = ((m1y - m4y) * m3x - (m1x - m4x) * m3y + m1x * m4y - m1y * m4x) / (
        (m2y - m4y) * m3x - (m2x - m4x) * m3y + m2x * m4y - m2y * m4x
    )
    k3 = ((m1y - m4y) * m2x - (m1x - m4x) * m2y + m1x * m4y - m1y * m4x) / (
        (m3y - m4y) * m2x - (m3x - m4x) * m2y + m3x * m4y - m3y * m4x
    )

    # f_squared is the focal length of the camera, squared
    # if k2==1 OR k3==1 then this equation is not solvable
    # if the focal length is known, then this equation is not needed
    # in that case assign f_squared= sqr(focal_length)
    if focal_length:
        f_squared = sqr(focal_length)
    else:
        f_squared = -((k3 * m3y - m1y) * (k2 * m2y - m1y) + (k3 * m3x - m1x) * (k2 * m2x - m1x)) / ((k3 - 1) * (k2 - 1))

    try:
        logging.info(f"Focal length: {sqrt(f_squared)}")
    except ValueError:
        logging.warning(
            f"estimating FL failed, {[(m1x, m1y), (m2x, m2y), (m3x, m3y), (m4x, m4y)]}, k2: {k2}, k3: {k3}, f^2: {f_squared}"
        )

    try:
        # The width/height ratio of the original rectangle
        whRatio = sqrt(
            (sqr(k2 - 1) + sqr(k2 * m2y - m1y) / f_squared + sqr(k2 * m2x - m1x) / f_squared)
            / (sqr(k3 - 1) + sqr(k3 * m3y - m1y) / f_squared + sqr(k3 * m3x - m1x) / f_squared)
        )
    except ValueError:
        logging.warning(
            f"estimating AR failed, {[(m1x, m1y), (m2x, m2y), (m3x, m3y), (m4x, m4y)]}, k2: {k2}, k3: {k3}, f^2: {f_squared}"
        )
        return -1

    # if k2==1 AND k3==1, then the focal length equation is not solvable
    # but the focal length is not needed to calculate the ratio.
    # I am still trying to figure out under which circumstances k2 and k3 become 1
    # but it seems to be when the rectangle is not distorted by perspective,
    # i.e. viewed straight on. Then the equation is obvious:
    if k2 == 1 and k3 == 1:
        whRatio = sqrt((sqr(m2y - m1y) + sqr(m2x - m1x)) / (sqr(m3y - m1y) + sqr(m3x - m1x)))

    # After testing, I found that the above equations
    # actually give the height/width ratio of the rectangle,
    # not the width/height ratio.
    # If someone can find the error that caused this,
    # I would be most grateful.
    # until then:
    return whRatio
    whRatio = 1 / whRatio



from itertools import chain
from typing import TYPE_CHECKING

from networkx import Graph, connected_components

if TYPE_CHECKING:
    from typing import Sequence, TypeVar

    import numpy as np

    T = TypeVar("T")


def complete_weighted_bipartite_graph(first: Sequence[T], second: Sequence[T], weights: np.ndarray) -> Graph:
    graph = Graph()

    graph.add_nodes_from(first, bipartite=0)
    graph.add_nodes_from(second, bipartite=1)

    for i in range(len(first)):
        for j in range(len(second)):
            graph.add_edge(first[i], second[j], weight=weights[i][j])

    return graph


def connected_subgraph(graph: Graph, minsize: int = 2) -> Graph:
    """Returns all the connected subgraphs of `graph`,
    which are at least of size `minsize`.
    """

    nodelist = list(chain.from_iterable(comp for comp in connected_components(graph) if len(comp) >= minsize))
    return graph.subgraph(nodelist)



from datetime import date, datetime, time
from functools import partial
from typing import Any, Iterable, Iterator

from msgpack import ExtType, Packer, Unpacker, pack, packb, unpack, unpackb

from ._files import PathType
from .atomic import sopen
from .file import copen


class EXTID:
    datetime = 1
    set = 2
    frozenset = 3
    date = 4
    time = 5


def default(obj):
    if isinstance(obj, datetime):
        return ExtType(EXTID.datetime, obj.isoformat().encode("utf-8"))  # todo: use struct.Struct instead
    elif isinstance(obj, date):
        return ExtType(EXTID.date, obj.isoformat().encode("utf-8"))  # todo: use struct.Struct instead
    elif isinstance(obj, time):
        return ExtType(EXTID.time, obj.isoformat().encode("utf-8"))  # todo: use struct.Struct instead
    elif isinstance(obj, set):
        return ExtType(EXTID.set, packb(tuple(obj), use_bin_type=True))
    elif isinstance(obj, frozenset):
        return ExtType(EXTID.frozenset, packb(tuple(obj), use_bin_type=True))

    return obj


def ext_hook(code, data):
    if code == EXTID.datetime:
        return datetime.fromisoformat(data.decode("utf-8"))
    elif code == EXTID.date:
        return date.fromisoformat(data.decode("utf-8"))
    elif code == EXTID.time:
        return time.fromisoformat(data.decode("utf-8"))
    elif code == EXTID.set:
        return set(unpackb(data, use_list=False, raw=False))
    elif code == EXTID.frozenset:
        return frozenset(unpackb(data, use_list=False, raw=False))

    return data


dumps = partial(packb, default=default, use_bin_type=True)
loads = partial(unpackb, use_list=False, raw=False, strict_map_key=False, ext_hook=ext_hook)


def key_to_hash(key: Any) -> str:
    from hashlib import md5

    binary = dumps(key)
    return md5(binary).hexdigest()  # nosec


def read_msgpack(path: PathType) -> Any:
    """Read msgpack file from `path`."""

    with copen(path, "rb") as fr:
        return unpack(fr, use_list=False, raw=False, strict_map_key=False, ext_hook=ext_hook)


def write_msgpack(obj: Any, path: PathType, safe: bool = False) -> None:
    """Write `obj` to `path` using msgpack serialization.

    `safe`: if True, don't overwrite original file in case any error occurs
    """

    with sopen(path, "wb", safe=safe) as fw:
        pack(obj, fw, default=default, use_bin_type=True)


def read_iter(path: PathType) -> Iterator[Any]:
    """Read msgpack'd iterable from `path`."""

    with copen(path, "rb") as fr:
        unpacker = Unpacker(fr, use_list=False, raw=False, strict_map_key=False, ext_hook=ext_hook)
        yield from unpacker


def write_iter(it: Iterable[Any], path: PathType, safe: bool = False) -> Iterator[Any]:
    """Write iterable `it` to `path` using msgpack serialization. This uses much less memory than
            writing a full list at once.
    Read back using `read_iter()`. If `safe` is True, the original file is not overwritten
            if any error occurs.
    This is a generator which yields the values read from `it`. So it must be consumed
            to actually write anything to disk.
    """

    with sopen(path, "wb", safe=safe) as fw:
        packer = Packer(default=default, use_bin_type=True)
        for obj in it:
            fw.write(packer.pack(obj))
            yield obj



import re
from collections import OrderedDict
from functools import partial
from itertools import chain
from locale import strxfrm
from typing import Any, Callable, Dict, Iterable, Iterator, List, Mapping, Optional, Sequence, Tuple, TypeVar, Union

from .binary import decode_binary, encode_binary
from .iter import switched_enumerate

T = TypeVar("T")
UnicodeOrdinalT = Union[int, str]

english_consonants = "bcdfghjklmnpqrstvwxz"  # y?
english_vowels = "aeiouy"  # y?

german_consonants = "bcdfghjklmnpqrstvwxyz"
german_vowels = "aeiou"

# def cycleleft(s, start=0, end=-1): # bad
#     return s[:start] + s[start+1:end] + s[start] + s[end+1:]

# def cycleright(s, start=0, end=-1): # bad
#     return s[:start] + s[end] + s[start:end] + s[end+1:]


def backslash_escaped_ascii(text: bytes) -> str:
    return text.decode("ascii", "backslashreplace")


def truncate(text: str, width: int, placeholder: str = "...") -> str:
    """Simple character based truncate method.
    To truncate at word boundaries use `textwrap.shorten`.
    """

    if width < 0:
        raise ValueError("width cannot be negative")

    if len(placeholder) > width:
        raise ValueError("placeholder cannot be longer then width")

    if len(text) <= width:
        return text
    else:
        return text[0 : width - len(placeholder)] + placeholder


def toint(obj: Any, default: T = None) -> Union[int, T]:
    """Converts `obj` to int if possible and returns `default` otherwise."""

    try:
        return int(obj)
    except (ValueError, TypeError):
        return default


def tryint(obj: T) -> Union[int, T]:
    """Converts `obj` to int if possible and returns the original object otherwise."""

    try:
        return int(obj)  # type: ignore
    except (ValueError, TypeError):
        return obj


def removeprefix(s: str, prefix: str) -> str:
    """If string `s` starts with `prefix` cut it off and return the remaining string."""

    if s.startswith(prefix):
        return s[len(prefix) :]
    else:
        return s


def removeprefixes(s: str, prefixes: Iterable[str]) -> str:
    """If string `s` starts with one of `prefixes` cut it off and return the remaining string."""

    for prefix in prefixes:
        if s.startswith(prefix):
            return s[len(prefix) :]

    return s


def removesuffix(s: str, suffix: str) -> str:
    """If string `s` ends with `suffix` cut it off and return the preceding string."""

    if s.endswith(suffix):
        return s[: -len(suffix)]
    else:
        return s


def removesuffixes(s: str, suffixes: Iterable[str]) -> str:
    """If string `s` ends with one of the `suffixes` cut it off and return the preceding string."""

    for suffix in suffixes:
        if s.endswith(suffix):
            return s[: -len(suffix)]

    return s


def encode_case(s: str) -> bytes:
    """Encode case information of a string to a bit-encoded bytes string."""

    return encode_binary(c.isupper() for c in s)


def decode_case(s: str, key: bytes) -> str:
    """Apply the case information encoded in `key` to string `s`."""

    return "".join(c.upper() if b else c for c, b in zip(s, decode_binary(key)))


def locale_sorted(seq: Sequence[str], case_insensitive: bool = True, lower_before_upper: bool = True) -> List[str]:
    if case_insensitive:
        if lower_before_upper:

            def key(s):
                return (strxfrm(s.lower()), strxfrm(s.swapcase()))

        else:

            def key(s):
                return (strxfrm(s.lower()), strxfrm(s))

    else:
        if lower_before_upper:

            def key(s):
                return (strxfrm(s.swapcase()), strxfrm(s))

        else:
            key = strxfrm

    return sorted(seq, key=key)


def build_multiple_replace(
    d: Mapping[str, str], escape: bool = True, ignorecase: bool = False, longestfirst: bool = False
) -> Callable[[str], str]:
    """Returns a callable, which when applied to a string,
    replaces all the keys in `d` with the corresponding values.
    The replacement happens in iteration order of the mapping.
    The complexity then is linear in  the length of the input string.
    """

    if escape:
        it: Iterable[str] = map(re.escape, d.keys())
    else:
        it = d.keys()

    if longestfirst:
        it = sorted(it, key=len, reverse=True)

    flags = re.IGNORECASE if ignorecase else 0
    cp = re.compile("(" + "|".join(it) + ")", flags)
    return partial(cp.sub, lambda m: d[m.group(0)])


def multiple_replace(d: Mapping[str, str], s: str, ignorecase: bool = False, longestfirst: bool = False) -> str:
    """Uses mapping `d` to replace keys with values in `s`.
    This is a wrapper for `build_multiple_replace`,
    which should be used directly for improved performance
    in case the same `d` is used for multiple `s`.
    """

    f = build_multiple_replace(d, ignorecase=ignorecase, longestfirst=longestfirst)
    return f(s)


def build_multiple_replace_list(
    old: Sequence[str], new: str, escape: bool = True, ignorecase: bool = False, longestfirst: bool = False
) -> Callable[[str], str]:
    """Returns a callable, which when applied to a string,
    replaces all the keys in `d` with the corresponding values.
    The replacement happens in iteration order of the mapping.
    The complexity then is linear in  the length of the input string.
    """

    if escape:
        it: Iterable[str] = map(re.escape, old)
    else:
        it = old

    if longestfirst:
        it = sorted(it, key=len, reverse=True)

    flags = re.IGNORECASE if ignorecase else 0
    cp = re.compile("(" + "|".join(it) + ")", flags)
    return partial(cp.sub, new)


def multiple_replace_list(
    old: Sequence[str], new: str, s: str, ignorecase: bool = False, longestfirst: bool = False
) -> str:
    """Uses sequence `old` to replace values in `s`.
    This is a wrapper for `build_multiple_replace_list`,
    which should be used directly for improved performance
    in case the same `old` is used for multiple `s`.
    """

    f = build_multiple_replace_list(old, new, escape=True, ignorecase=ignorecase, longestfirst=longestfirst)
    return f(s)


def replace_multiple(s: Iterable[str], d: Mapping[str, str]) -> str:  # cython candidate...
    """Use dictionary `d` to replace instances of key with value.
    If input `s` is a string, the keys must be strings of length 1.
    If the input is an iterable of strings,
    the strings will be compared to the keys for equality.
    The complexity is linear in the length of the iterable/string.
    For string inputs and keys of length > 1 see `multiple_replace`.
    """

    return "".join(d.get(c, c) for c in s)


def backslash_unescape(s: str) -> str:
    """Converts strings with backslash-escaped entities like \n or \u1234
    to a string with these entities.
    Example: backslash_unescape("\\n") -> "\n"
    """

    # for some reason unicode_escape is based on latin-1
    # ascii also works, but the intermittent string would be longer
    # because more things will be backslash-escaped
    return s.encode("latin-1", "backslashreplace").decode("unicode_escape")


_backslashquote_escape = build_multiple_replace(OrderedDict([("\\", "\\\\"), ('"', '\\"')]))


def backslashquote_escape(s: str) -> str:
    """Converts \\ to \\\\ and " to \\" """

    return _backslashquote_escape(s)


_backslashquote_unescape = build_multiple_replace(OrderedDict([('\\"', '"'), ("\\\\", "\\")]))


def backslashquote_unescape(s: str) -> str:
    """Unescapes \\ and \" in `s`."""

    return _backslashquote_unescape(s)


_backslashcontrol_escape = build_multiple_replace(
    OrderedDict([("\\", "\\\\"), ("\t", "\\t"), ("\n", "\\n"), ("\r", "\\r")])
)


def backslashcontrol_escape(s: str) -> str:
    """Escapes \\ and \\t \\n \\r in `s`."""

    return _backslashcontrol_escape(s)


_backslashcontrol_unescape = build_multiple_replace(
    OrderedDict([("\\\\", "\\"), ("\\t", "\t"), ("\\n", "\n"), ("\\r", "\r")])
)


def backslashcontrol_unescape(s: str) -> str:
    """Unescapes \\\\ and \\\\t \\\\n \\\\r in `s`."""

    return _backslashcontrol_unescape(s)


def are_parentheses_matched(s: str, open: str = "([{", close: str = ")]}") -> bool:
    stack = list()
    parentheses = dict(chain(switched_enumerate(open), switched_enumerate(close)))
    assert len(parentheses) == len(open) + len(close)

    for char in s:
        if char in open:
            stack.append(parentheses[char])
        if char in close:
            try:
                if stack.pop() != parentheses[char]:  # wrong nesting
                    return False
            except IndexError:  # stack empty
                return False

    return len(stack) == 0  # unclosed parentheses left


def filter_join(s: str, it: Iterable[str], func: Optional[Callable[[str], bool]] = None) -> str:
    if func is None:
        return s.join(i for i in it if i)
    else:
        return s.join(i for i in it if func(i))


def surrounding_join(j: str, it: Iterable[str], left: str = "", right: str = "") -> str:
    """Example: surrounding_join(", ", ("a", "b", "c"), left="[", right="]") -> "[a], [b], [c]" """

    s = (right + j + left).join(it)
    if s != "":
        return left + s + right
    return ""


def replace_pairs_bytes(s: bytes, items: Dict[bytes, Optional[bytes]]) -> bytes:
    frm = b"".join(k for k, v in items.items() if v)
    to = b"".join(v for k, v in items.items() if v)
    delete = b"".join(k for k, v in items.items() if v is None)

    table = s.maketrans(frm, to)

    return s.translate(table, delete)


def replace_pairs_chars(s: str, items: Dict[UnicodeOrdinalT, Optional[UnicodeOrdinalT]]) -> str:
    table = s.maketrans(items)
    return s.translate(table)


_contains_digit = re.compile(r"\d")


def contains_digit(s: str) -> bool:
    """Tests if a digit is contained in string `s`."""

    return _contains_digit.search(s) is not None


def all_splits(text: str, include_start: bool = False, include_end: bool = True) -> Iterator[Tuple[str, str]]:
    if include_start:
        start = 0
    else:
        start = 1

    if include_end:
        end = 1
    else:
        end = 0

    return ((text[:i], text[i:]) for i in range(start, len(text) + end))



from itertools import cycle
from typing import Iterable, Iterator


def xor_stream(it: Iterable[bytes], key: Iterable[bytes], cyclekey: bool = True) -> Iterator[bytes]:
    if cyclekey:
        key = cycle(key)

    for i, j in zip(it, key):
        yield bytes([i ^ j])



import sys
from types import FrameType
from typing import Any, Optional

""" All functions of this module use implemention details specific to CPython
    and might not work on other implementations.
"""


def find_local_in_call_stack(funcname: str, varname: str, stacklevel: int = 2) -> Any:
    """Finds a local variable by name `varname` by going up the call stack,
    looking for functions called `funcname`.

    The functions starts by looking at the calling frame of the calling point of this function by default.
    It can start a different point in the stack be changing `stacklevel`.
    Raises a KeyError if the variable cannot be found.

    Example:
            def func_a():
                    var_a = "asd"
                    func_b()

            def func_b():
                    # starts by looking at func_a and then going further up the stack
                    var_a = find_local_in_call_stack("func_a", "var_a")
                    assert var_a == "asd"

            func_a()
    """

    frame: Optional[FrameType] = sys._getframe(stacklevel)
    while frame:
        if frame.f_code.co_name == funcname and varname in frame.f_locals:
            return frame.f_locals[varname]
        frame = frame.f_back

    raise KeyError(f"Did not find variable {varname} in function {funcname}")



import os.path
from abc import ABCMeta, abstractmethod
from typing import Any, Coroutine, Dict, Generic, List, Optional, TypeVar, Union

import aiohttp
import requests

from .exceptions import assert_choice
from .yaml import read_yaml

""" Properly type checking this file would require higher-kinded type-vars,
    ie. https://github.com/python/typing/issues/548 to be fixed.
"""

JsonDict = Dict[str, Any]
JsonValue = Union[JsonDict, List[JsonDict]]
ReturnT = TypeVar("ReturnT")  # fixme: should be higher kinded type var
ResponseT = Union[bytes, JsonValue]

SyncReturnT = ResponseT
AsyncReturnT = Coroutine[Any, Any, ResponseT]


class RasaABC(Generic[ReturnT], metaclass=ABCMeta):
    sender: str

    @abstractmethod
    def get_endpoint(self, path: str) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_request(self, url: str, params: Optional[JsonDict] = None, raw: bool = False) -> ReturnT:
        raise NotImplementedError

    @abstractmethod
    def post_request(
        self, url: str, params: Optional[JsonDict] = None, json: Optional[JsonDict] = None, raw: bool = False
    ) -> ReturnT:
        raise NotImplementedError

    @abstractmethod
    def put_request(
        self, url: str, params: Optional[JsonDict] = None, json: Optional[JsonDict] = None, raw: bool = False
    ) -> ReturnT:
        raise NotImplementedError

    @abstractmethod
    def delete_request(
        self, url: str, params: Optional[JsonDict] = None, json: Optional[JsonDict] = None, raw: bool = False
    ) -> ReturnT:
        raise NotImplementedError


class Rasa:
    sender: str

    def __init__(
        self,
        sender: str,
        scheme: str = "http",
        netloc: str = "localhost:5005",
        token: Optional[str] = None,
        timeout: int = 60,
    ) -> None:
        self.sender = sender
        self.scheme = scheme
        self.netloc = netloc
        self.token = token
        self.timeout = timeout

    def get_endpoint(self, path: str) -> str:
        return f"{self.scheme}://{self.netloc}{path}"


class RasaRest(Rasa):
    def get_request(self, url: str, params: Optional[JsonDict] = None, raw: bool = False) -> ResponseT:
        params = params or {}

        if self.token:
            params.setdefault("token", self.token)

        r = requests.get(url, timeout=self.timeout, params=params)
        r.raise_for_status()
        if raw:
            return r.content
        else:
            return r.json()

    def post_request(
        self, url: str, params: Optional[JsonDict] = None, json: Optional[JsonDict] = None, raw: bool = False
    ) -> ResponseT:
        params = params or {}

        if self.token:
            params.setdefault("token", self.token)

        r = requests.post(url, timeout=self.timeout, params=params, json=json)
        r.raise_for_status()
        if raw:
            return r.content
        else:
            return r.json()

    def put_request(
        self, url: str, params: Optional[JsonDict] = None, json: Optional[JsonDict] = None, raw: bool = False
    ) -> ResponseT:
        params = params or {}

        if self.token:
            params.setdefault("token", self.token)

        r = requests.put(url, timeout=self.timeout, params=params, json=json)
        r.raise_for_status()

        if raw:
            return r.content
        else:
            if r.status_code == 204:
                return b""
            else:
                return r.json()

    def delete_request(
        self, url: str, params: Optional[JsonDict] = None, json: Optional[JsonDict] = None, raw: bool = False
    ) -> ResponseT:
        params = params or {}

        if self.token:
            params.setdefault("token", self.token)

        r = requests.delete(url, timeout=self.timeout, params=params, json=json)
        r.raise_for_status()

        if raw:
            return r.content
        else:
            if r.status_code == 204:
                return b""
            else:
                return r.json()


class RasaRestAsync(Rasa):
    async def get_request(self, url: str, params: Optional[JsonDict] = None, raw: bool = False) -> ResponseT:
        params = params or {}

        if self.token:
            params.setdefault("token", self.token)

        async with aiohttp.ClientSession() as session:
            async with session.get(url, timeout=self.timeout, params=params) as r:
                r.raise_for_status()
                if raw:
                    return await r.read()
                else:
                    return await r.json()

    async def post_request(
        self, url: str, params: Optional[JsonDict] = None, json: Optional[JsonDict] = None, raw: bool = False
    ) -> ResponseT:
        params = params or {}

        if self.token:
            params.setdefault("token", self.token)

        async with aiohttp.ClientSession() as session:
            async with session.post(url, timeout=self.timeout, params=params, json=json) as r:
                r.raise_for_status()
                if raw:
                    return await r.read()
                else:
                    return await r.json()

    async def put_request(
        self, url: str, params: Optional[JsonDict] = None, json: Optional[JsonDict] = None, raw: bool = False
    ) -> ResponseT:
        params = params or {}

        if self.token:
            params.setdefault("token", self.token)

        async with aiohttp.ClientSession() as session:
            async with session.put(url, timeout=self.timeout, params=params, json=json) as r:
                r.raise_for_status()
                if raw:
                    return await r.read()
                else:
                    if r.status == 204:
                        return {}
                    else:
                        return await r.json()

    async def delete_request(
        self, url: str, params: Optional[JsonDict] = None, json: Optional[JsonDict] = None, raw: bool = False
    ) -> ResponseT:
        params = params or {}

        if self.token:
            params.setdefault("token", self.token)

        async with aiohttp.ClientSession() as session:
            async with session.delete(url, timeout=self.timeout, params=params, json=json) as r:
                r.raise_for_status()
                if raw:
                    return await r.read()
                else:
                    if r.status == 204:
                        return {}
                    else:
                        return await r.json()


INCLUDE_EVENTS_ENUM = {"AFTER_RESTART", "ALL", "APPLIED", "NONE"}
OUTPUT_CHANNEL_ENUM = {
    "latest",
    "slack",
    "callback",
    "facebook",
    "rocketchat",
    "telegram",
    "twilio",
    "webexteams",
    "socketio",
}


class _RasaRestConversations(RasaABC[ReturnT]):
    def get_tracker(self, include_events: Optional[str] = None, until: Optional[int] = None) -> ReturnT:
        """Retrieve a conversations tracker.

        The tracker represents the state of the conversation. The state of the tracker is created by applying a sequence of events, which modify the state. These events can optionally be included in the response.
        """

        assert_choice("include_events", include_events, INCLUDE_EVENTS_ENUM, True)

        url = self.get_endpoint(f"/conversations/{self.sender}/tracker")

        return self.get_request(
            url,
            params={
                "include_events": include_events,
                "until": until,
            },
        )

    def post_events(
        self,
        event: str,
        timestamp: int,
        include_events: Optional[str] = None,
        output_channel: Optional[str] = None,
        execute_side_effects: bool = False,
    ) -> ReturnT:
        """Append events to a tracker.

        Appends one or multiple new events to the tracker state of the conversation. Any existing events will be kept and the new events will be appended, updating the existing state. If events are appended to a new conversation ID, the tracker will be initialised with a new session.
        """

        assert_choice("include_events", include_events, INCLUDE_EVENTS_ENUM, True)
        assert_choice("output_channel", output_channel, OUTPUT_CHANNEL_ENUM, True)

        url = self.get_endpoint(f"/conversations/{self.sender}/tracker/events")

        return self.post_request(
            url,
            params={
                "include_events": include_events,
                "output_channel": output_channel,
                "execute_side_effects": execute_side_effects,
            },
            json={
                "event": event,
                "timestamp": timestamp,
            },
        )

    def get_story(self, until: Optional[int] = None, all_sessions: bool = False) -> ReturnT:
        """Retrieve an end-to-end story corresponding to a conversation.

        The story represents the whole conversation in end-to-end format. This can be posted to the '/test/stories' endpoint and used as a test.
        """

        url = self.get_endpoint(f"/conversations/{self.sender}/story")

        return self.get_request(
            url,
            params={
                "until": until,
                "all_sessions": all_sessions,
            },
            raw=True,
        )

    def execute_action(
        self,
        name: str,
        policy: Optional[str] = None,
        confidence: Optional[float] = None,
        include_events: Optional[str] = None,
        output_channel: Optional[str] = None,
    ) -> ReturnT:
        """Run an action in a conversation.

        DEPRECATED. Runs the action, calling the action server if necessary. Any responses sent by the executed action will be forwarded to the channel specified in the `output_channel` parameter. If no output channel is specified, any messages that should be sent to the user will be included in the response of this endpoint.
        """

        assert_choice("include_events", include_events, INCLUDE_EVENTS_ENUM, True)
        assert_choice("output_channel", output_channel, OUTPUT_CHANNEL_ENUM, True)

        url = self.get_endpoint(f"/conversations/{self.sender}/execute")

        return self.post_request(
            url,
            params={
                "include_events": include_events,
                "output_channel": output_channel,
            },
            json={
                "name": name,
                "policy": policy,
                "confidence": confidence,
            },
        )

    def trigger_intent(
        self,
        name: str,
        entities: Optional[JsonDict] = None,
        include_events: Optional[str] = None,
        output_channel: Optional[str] = None,
    ) -> ReturnT:
        """Inject an intent into a conversation.

        Sends a specified intent and list of entities in place of a user message. The bot then predicts and executes a response action. Any responses sent by the executed action will be forwarded to the channel specified in the output_channel parameter. If no output channel is specified, any messages that should be sent to the user will be included in the response of this endpoint.
        """

        assert_choice("include_events", include_events, INCLUDE_EVENTS_ENUM, True)
        assert_choice("output_channel", output_channel, OUTPUT_CHANNEL_ENUM, True)

        url = self.get_endpoint(f"/conversations/{self.sender}/trigger_intent")

        return self.post_request(
            url,
            params={
                "include_events": include_events,
                "output_channel": output_channel,
            },
            json={
                "name": name,
                "entities": entities,
            },
        )


REMOTE_STORAGE_ENUM = {"aws", "gcs", "azure"}


class _RasaRestModel(RasaABC[ReturnT]):
    def train(
        self,
        directory: str,
        config: Optional[str] = None,
        domain: Optional[str] = None,
        nlu: Optional[str] = None,
        responses: Optional[str] = None,
        stories: Optional[str] = None,
        save_to_default_model_directory: bool = True,
        force_training: bool = False,
    ) -> ReturnT:
        """Train a Rasa model.

        Trains a new Rasa model. Depending on the data given only a dialogue model, only a NLU model, or a model combining a trained dialogue model with an NLU model will be trained. The new model is not loaded by default.
        """

        # use default values
        json = {
            "config": config or "config.yml",
            "domain": domain or "domain.yml",
            "nlu": nlu or "data/nlu.yml",
            "responses": responses or "data/responses.yml",
            "stories": stories or "data/stories.yml",
        }

        # check if files exist and read contents
        json = {k: read_yaml(v) if os.path.exists(v) else None for k, v in json.items()}

        url = self.get_endpoint("/model/train")

        return self.post_request(
            url,
            params={
                "save_to_default_model_directory": save_to_default_model_directory,
                "force_training": force_training,
            },
            json=json,
        )

    def replace_model(
        self,
        model_file: Optional[str] = None,
        model_server: Optional[JsonDict] = None,
        remote_storage: Optional[str] = None,
    ) -> ReturnT:
        """Replace the currently loaded model.

        Updates the currently loaded model. First, tries to load the model from the local storage system. Secondly, tries to load the model from the provided model server configuration. Last, tries to load the model from the provided remote storage.
        """

        assert_choice("remote_storage", remote_storage, REMOTE_STORAGE_ENUM, True)

        if model_file is None and model_server is None and remote_storage is None:
            model_file = "models/"

        url = self.get_endpoint("/model")

        return self.put_request(
            url,
            json={
                "model_file": model_file,
                "model_server": model_server,
                "remote_storage": remote_storage,
            },
        )

    def unload_model(self) -> ReturnT:
        url = self.get_endpoint("/model")

        return self.delete_request(url)


class _RasaRestWebhook(RasaABC[ReturnT]):
    def health(self) -> ReturnT:
        url = self.get_endpoint("/webhooks/rest/")

        return self.get_request(url)

    def send_message(self, message: str) -> ReturnT:
        url = self.get_endpoint("/webhooks/rest/webhook")

        return self.post_request(
            url,
            json={
                "sender": self.sender,
                "message": message,
            },
        )


class _RasaCallbackWebhook(RasaABC[ReturnT]):
    def send_message(self, message: str) -> ReturnT:
        url = self.get_endpoint("/webhooks/callback/webhook")

        return self.post_request(
            url,
            json={
                "sender": self.sender,
                "message": message,
            },
        )


class RasaRestConversations(RasaRest, _RasaRestConversations[SyncReturnT]):
    pass


class RasaRestModel(RasaRest, _RasaRestModel[SyncReturnT]):
    pass


class RasaRestWebhook(RasaRest, _RasaRestWebhook[SyncReturnT]):
    pass


class RasaCallbackWebhook(RasaRest, _RasaCallbackWebhook[SyncReturnT]):
    pass


class RasaRestConversationsAsync(RasaRestAsync, _RasaRestConversations[AsyncReturnT]):
    pass


class RasaRestModelAsync(RasaRestAsync, _RasaRestModel[AsyncReturnT]):
    pass


class RasaRestWebhookAsync(RasaRestAsync, _RasaRestWebhook[AsyncReturnT]):
    pass


class RasaCallbackWebhookAsync(RasaRestAsync, _RasaCallbackWebhook[AsyncReturnT]):
    pass



import os
import platform
import sys
from pathlib import Path
from typing import Callable, Union

from typing_extensions import Self

from ._func import rename
from .os_shared import is_os_64bit  # noqa: F401

PathStr = Union[Path, str]

system = platform.system()


class CurrentWorkingDirectory:
    __slots__ = ("oldcwd",)

    def __init__(self, path: PathStr) -> None:
        self.oldcwd = os.getcwd()
        os.chdir(path)

    def close(self) -> None:
        os.chdir(self.oldcwd)

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


def _not_available(func_name: str) -> Callable:
    @rename(func_name)
    def inner(*args, **kwargs):
        raise OSError(f"{func_name}() is not available for {system}")

    return inner


if system == "Windows":
    from .os_win import _disk_usage_windows as disk_usage
    from .os_win import _filemanager_cmd_windows as filemanager_cmd
    from .os_win import _get_appdata_dir as get_appdata_dir
    from .os_win import _interrupt_windows as interrupt
    from .os_win import _islink as islink
    from .os_win import _lock as lock
    from .os_win import _unlock as unlock
    from .os_win import _volume_info_windows as volume_info

    if sys.version_info >= (3, 8):
        from os.path import realpath
    else:
        from .os_win import _realpath as realpath

elif system == "Linux":
    from .os_posix import _disk_usage_posix as disk_usage
    from .os_posix import _get_appdata_dir as get_appdata_dir
    from .os_posix import _lock as lock
    from .os_posix import _unlock as unlock

    volume_info = _not_available("volume_info")
    from os.path import islink, realpath

    from .os_posix import _filemanager_cmd_posix as filemanager_cmd
    from .os_posix import _interrupt_posix as interrupt

elif system == "Darwin":
    from .os_posix import _disk_usage_posix as disk_usage
    from .os_posix import _lock as lock
    from .os_posix import _unlock as unlock

    volume_info = _not_available("volume_info")
    from .os_mac import _filemanager_cmd_mac as filemanager_cmd

    get_appdata_dir = _not_available("get_appdata_dir")
    from os.path import islink, realpath

    from .os_posix import _interrupt_posix as interrupt

else:
    lock = _not_available("lock")
    unlock = _not_available("unlock")
    from shutil import disk_usage

    volume_info = _not_available("volume_info")
    filemanager_cmd = _not_available("filemanager_cmd")
    get_appdata_dir = _not_available("get_appdata_dir")
    interrupt = _not_available("interrupt")
    from os.path import islink, realpath

lock.__doc__ = """ Locks access to the file (on Posix) or its contents (Windows). """
unlock.__doc__ = """ Unlocks access to the file. """
disk_usage.__doc__ = """ Returns (total, used, free) bytes on disk. """
volume_info.__doc__ = """ filesystem and name of the volume """
filemanager_cmd.__doc__ = """ Returns a shell command that when executed starts the file manager of the OS. """
get_appdata_dir.__doc__ = """ Returns the roaming appdata directory of the current user. """
islink.__doc__ = """ islink """
realpath.__doc__ = """ realpath """
interrupt.__doc__ = """ interrupt """



import os.path
from os import PathLike, fspath, remove, replace
from tempfile import mkstemp
from typing import IO, ContextManager, Optional, Union

from .file import copen

PathType = Union[str, PathLike]


# http://stupidpythonideas.blogspot.tw/2014/07/getting-atomic-writes-right.html
class TransactionalCreateFile:
    def __init__(
        self,
        path: PathType,
        mode: str = "wb",
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        newline: Optional[str] = None,
        prefix: str = "tmp",
        handle_archives: bool = True,
    ) -> None:
        is_text = "t" in mode

        self.path = fspath(path)
        suffix = os.path.splitext(self.path)[1].lower()
        curdir = os.path.dirname(self.path)
        fd, self.tmppath = mkstemp(suffix, prefix, curdir, is_text)
        self.fp = copen(
            fd, mode, encoding=encoding, errors=errors, newline=newline, ext=suffix, handle_archives=handle_archives
        )

    def commit(self) -> None:
        self.fp.close()
        replace(self.tmppath, self.path)  # should be atomic

    def rollback(self) -> None:
        self.fp.close()
        remove(self.tmppath)

    def __enter__(self) -> IO:
        return self.fp

    def __exit__(self, exc_type, exc_value, traceback):
        # at this point the original file is unmodified and the new file exists as tempfile on disk (or in buffer on windows)
        if exc_type:
            self.rollback()
        else:
            self.commit()


def sopen(
    path: PathType,
    mode: str = "rb",
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    newline: Optional[str] = None,
    safe: bool = False,
) -> ContextManager[IO]:
    if safe:
        return TransactionalCreateFile(path, mode, encoding=encoding, errors=errors, newline=newline)
    else:
        return copen(path, mode, encoding=encoding, errors=errors, newline=newline)


def write_file(
    data: Union[str, bytes],
    path: PathType,
    mode: str = "wb",
    encoding: Optional[str] = None,
    errors: Optional[str] = None,
    newline: Optional[str] = None,
) -> None:
    """Writes/overwrites files in a safe way. That means either the original file
    will be left untouched, or it will be replaced with the complete new file.
    """

    with TransactionalCreateFile(path, mode, encoding=encoding, errors=errors, newline=newline) as fw:
        fw.write(data)



import csv
import datetime
import json
import logging
from functools import partial
from itertools import islice
from pathlib import Path
from traceback import TracebackException
from types import ModuleType
from typing import IO, Any, Callable, Dict, FrozenSet, Iterable, Iterator, Optional, Sequence, Tuple, Type, Union

from typing_extensions import Self  # typing.Self is available in Python 3.11+
from typing_extensions import TypedDict  # typing.TypedDict is available in Python 3.8+

from .atomic import sopen
from .datetime import datetime_from_utc_timestamp_ms, now
from .file import copen

PathStr = Union[Path, str]
JsonDict = Dict[str, Any]

logger = logging.getLogger(__name__)


class BuiltinEncoder(json.JSONEncoder):
    def __init__(self, *args: Any, sort_sets: bool = False, **kwargs: Any) -> None:
        json.JSONEncoder.__init__(self, *args, **kwargs)
        self.sort_sets = sort_sets

    def default(self, obj):
        # collections.OrderedDict is supported by default

        from base64 import b85encode
        from uuid import UUID

        if isinstance(obj, (set, frozenset)):
            if self.sort_sets:
                return sorted(obj)
            else:
                return tuple(obj)
        elif isinstance(obj, complex):
            return [obj.real, obj.imag]
        elif isinstance(obj, (datetime.date, datetime.time, datetime.datetime)):
            return obj.isoformat()
        elif isinstance(obj, datetime.timedelta):
            return obj.total_seconds()
        elif isinstance(obj, UUID):
            return str(obj)
        elif isinstance(obj, bytes):
            return b85encode(obj).decode("ascii")  # b85encode doesn't use ", ' or \

        return json.JSONEncoder.default(self, obj)


class BuiltinRoundtripEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime.timedelta):
            return {"$timedelta": (obj.days, obj.seconds, obj.microseconds)}
        elif isinstance(obj, datetime.datetime):
            return {"$datetime": obj.isoformat()}
        elif isinstance(obj, datetime.date):
            return {"$dateobj": obj.isoformat()}
        elif isinstance(obj, set):
            return {"$set": tuple(obj)}
        elif isinstance(obj, frozenset):
            return {"$frozenset": tuple(obj)}

        return json.JSONEncoder.default(self, obj)


class BuiltinRoundtripDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, *args, object_hook=self.object_hook, **kwargs)

    def object_hook(self, obj):
        if len(obj) == 1:
            key, value = next(iter(obj.items()))
            if key == "$timedelta":
                return datetime.timedelta(*value)
            elif key == "$datetime":
                return datetime.datetime.fromisoformat(value)
            elif key == "$date":  # for compatibility with bson.json_util
                return datetime_from_utc_timestamp_ms(value)
            elif key == "$dateobj":
                return datetime.date.fromisoformat(value)
            elif key == "$set":
                return set(value)
            elif key == "$frozenset":
                return frozenset(value)

        return obj


class TracebackExceptionEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, TracebackException):
            return "\n".join(obj.format())
        return super().default(obj)


def read_json_schema(path: PathStr) -> JsonDict:
    with open(path, encoding="utf-8") as fr:
        return json.load(fr)


def read_json(
    path: PathStr,
    schema: Optional[Union[str, JsonDict]] = None,
    cls: Optional[Type[json.JSONDecoder]] = None,
    object_hook: Any = None,
) -> Any:
    """Read the json file at `path` and optionally validates the input according to `schema`.
    The validation requires `jsonschema`.
    `schema` can either be a path as well, or a Python dict which represents the schema.
    `cls` and `object_hook` is passed through to `json.load`.
    """

    with copen(path, "rt", encoding="utf-8") as fr:
        obj = json.load(fr, cls=cls, object_hook=object_hook)

    if schema is None:
        return obj

    from jsonschema import validate

    if isinstance(schema, str):
        schema = read_json_schema(schema)

    validate(obj, schema)
    return obj


def write_json(
    obj: Any,
    path: PathStr,
    mode: str = "wt",
    schema: Optional[Union[str, JsonDict]] = None,
    ensure_ascii: bool = False,
    cls: Optional[Type[json.JSONEncoder]] = None,
    indent: Optional[Union[str, int]] = None,
    sort_keys: bool = False,
    default: Optional[Callable] = None,
    safe: bool = False,
    **kw: Any,
) -> None:
    """Writes python object `obj` to `path` as json files and optionally validates the object
    according to `schema`. The validation requires `jsonschema`.
    The remaining optional parameters are passed through to `json.dump`.
    `safe`: if True, don't overwrite original file in case any error occurs
    """

    if schema:
        from jsonschema import validate

        if isinstance(schema, str):
            schema = read_json_schema(schema)

        validate(obj, schema)

    with sopen(path, mode, encoding="utf-8", safe=safe) as fw:
        json.dump(
            obj, fw, ensure_ascii=ensure_ascii, cls=cls, indent=indent, sort_keys=sort_keys, default=default, **kw
        )


class JsonLoadKwargs(TypedDict):
    cls: Optional[Type[json.JSONDecoder]]
    object_hook: Optional[Callable]
    parse_float: Optional[Callable]
    parse_int: Optional[Callable]
    parse_constant: Optional[Callable]
    object_pairs_hook: Optional[Callable]


class json_lines:
    """Read and write files in the JSON Lines format (http://jsonlines.org)."""

    def __init__(
        self,
        stream: IO,
        doclose: bool,
        cls: Optional[Type[json.JSONDecoder]] = None,
        object_hook: Optional[Callable] = None,
        parse_float: Optional[Callable] = None,
        parse_int: Optional[Callable] = None,
        parse_constant: Optional[Callable] = None,
        object_pairs_hook: Optional[Callable] = None,
        **kw: Any,
    ) -> None:
        """Don't use directly. Use `from_path` or `from_stream` classmethods instead."""

        """ fixme: how should `close` be handled?
            1: If you don't want to close `stream`, just don't call `close()` or use as context manager.
            2: use `doclose` argument to decide
        """

        self.f = stream
        self.doclose = doclose
        self.newline = "\n"

        self.json_kwargs: JsonLoadKwargs = {
            "cls": cls,
            "object_hook": object_hook,
            "parse_float": parse_float,
            "parse_int": parse_int,
            "parse_constant": parse_constant,
            "object_pairs_hook": object_pairs_hook,
        }
        self.json_cls_kw = kw

    @staticmethod
    def from_path(
        file: PathStr,
        mode: str = "rt",
        encoding: str = "utf-8",
        errors: str = "strict",
        newline: Optional[str] = None,
        cls: Optional[Type[json.JSONDecoder]] = None,
        object_hook: Optional[Callable] = None,
        parse_float: Optional[Callable] = None,
        parse_int: Optional[Callable] = None,
        parse_constant: Optional[Callable] = None,
        object_pairs_hook: Optional[Callable] = None,
        **kw: Any,
    ) -> "json_lines":
        """Binary writing modes "wb", "ab", "r+b", "w+b" are not supported by the python json module.
        Use text modes instead. Binary read-only is fine.
        """

        if isinstance(file, str):
            if file.startswith("s3://"):
                message = f"""Reading data from S3 is not supported natively. Use `smart-open` package:
```python
from smart_open import open
with open("{file}", "r") as fr:
    with json_lines.from_stream(fr) as jl:
        for doc in jl:
            pass
```"""
                raise ValueError(message)

        if set(mode) not in (set("rt"), set("rb"), set("xt"), set("xb"), set("wt"), set("at"), set("r+t"), set("w+t")):
            raise ValueError(f"mode cannot be `{mode}`")

        stream = copen(file, mode, encoding=encoding, errors=errors, newline=newline)
        return json_lines(
            stream, True, cls, object_hook, parse_float, parse_int, parse_constant, object_pairs_hook, **kw
        )

    @staticmethod
    def from_stream(
        stream: IO,
        cls: Optional[Type[json.JSONDecoder]] = None,
        object_hook: Optional[Callable] = None,
        parse_float: Optional[Callable] = None,
        parse_int: Optional[Callable] = None,
        parse_constant: Optional[Callable] = None,
        object_pairs_hook: Optional[Callable] = None,
        **kw: Any,
    ) -> "json_lines":
        return json_lines(
            stream, False, cls, object_hook, parse_float, parse_int, parse_constant, object_pairs_hook, **kw
        )

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def iterrange(self, start: int = 0, stop: Optional[int] = None) -> Iterator:
        try:
            import simplejson as sjson

            _json: ModuleType = sjson
        except ImportError:
            _json = json

        linenum = start + 1
        try:
            for line in islice(self.f, start, stop):
                line = line.rstrip().lstrip("\x00")  # fixme: strip \0 is only a temp fix!
                if line:
                    yield _json.loads(line, **self.json_kwargs, **self.json_cls_kw)
                linenum += 1

        except json.JSONDecodeError as e:
            e.lineno = linenum
            logger.error("JSON Lines parse error in line %s: '%r'", linenum, line)
            raise

    def __iter__(self) -> Iterator:
        return self.iterrange()

    def write(
        self,
        obj: Any,
        skipkeys: bool = False,
        ensure_ascii: bool = False,
        check_circular: bool = True,
        allow_nan: bool = True,
        cls: Optional[Type[json.JSONEncoder]] = None,
        separators: Optional[Tuple[str, str]] = None,
        default: Optional[Callable] = None,
        sort_keys: bool = False,
        **kw: Any,
    ) -> None:
        json.dump(
            obj,
            self.f,
            skipkeys=skipkeys,
            ensure_ascii=ensure_ascii,
            check_circular=check_circular,
            allow_nan=allow_nan,
            cls=cls,
            indent=None,
            separators=separators,
            default=default,
            sort_keys=sort_keys,
            **kw,
        )
        self.f.write(self.newline)

    def writelines(
        self,
        objs: Iterable[Any],
        skipkeys: bool = False,
        ensure_ascii: bool = False,
        check_circular: bool = True,
        allow_nan: bool = True,
        cls: Optional[Type[json.JSONEncoder]] = None,
        separators: Optional[Tuple[str, str]] = None,
        default: Optional[Callable] = None,
        sort_keys: bool = False,
        **kw: Any,
    ) -> None:
        for obj in objs:
            self.write(obj)

    def close(self) -> None:
        if self.doclose:
            self.f.close()

    def flush(self) -> None:
        self.f.flush()


def read_json_lines(
    file: PathStr, cls: Optional[Type[json.JSONDecoder]] = None, object_hook: Optional[Callable] = None
) -> Iterator[Any]:
    """Iterate over a JSON Lines `file` object by object.
    `object_hook` is passed through to `json.load`.
    """

    with json_lines.from_path(file, mode="rt", cls=cls, object_hook=object_hook) as fr:
        yield from fr


def write_json_lines(
    it: Iterable[Any],
    path,
    ensure_ascii: bool = False,
    sort_keys: bool = False,
    default: Optional[Callable] = None,
    safe: bool = False,
) -> Iterator[Any]:
    with sopen(path, "wt", encoding="utf-8", safe=safe) as fw:
        with json_lines.from_stream(fw, ensure_ascii=ensure_ascii, sort_keys=sort_keys, default=default) as fw:
            for obj in it:
                fw.write(obj)
                yield obj


def jl_to_csv(jlpath: PathStr, csvpath: str, keyfunc: Callable[[JsonDict], Sequence[str]], mode: str = "xt") -> None:
    with json_lines.from_path(jlpath, "rt") as fr:
        with open(csvpath, mode, encoding="utf-8", newline="") as csvfile:
            fw = csv.writer(csvfile)
            for obj in fr:
                fw.writerow(keyfunc(obj))


def key_to_hash(
    key: Any, ensure_ascii: bool = False, sort_keys: bool = False, default: Optional[Callable] = None
) -> str:
    from hashlib import md5

    binary = json.dumps(key, ensure_ascii=ensure_ascii, sort_keys=sort_keys, default=default).encode("utf-8")
    return md5(binary).hexdigest()  # nosec


class JsonLinesFormatter(logging.Formatter):
    """A JSON Lines formatter for the Python logging library.
    It expects a `dict` as logging message.
    For example:
            logger = logging.getLogger("jsonlines-test")
            handler = logging.StreamHandler()
            formatter = JsonLinesFormatter()
            handler.setFormatter(formatter)
            logger.addHandler(handler)
            logger.warning({"msg": "Hello world!", "level": "greeting"})
    """

    myfields = {
        "datetime": now,  # requires non-default json serializer
        "datetime-str": lambda: now().isoformat(),  # note: lambda required
    }

    def __init__(
        self,
        include: FrozenSet[str] = frozenset(),
        builtins: FrozenSet[str] = frozenset(),
        cls: Optional[Type[json.JSONEncoder]] = None,
        default: Optional[Callable] = None,
        sort_keys: bool = False,
    ) -> None:
        logging.Formatter.__init__(self)
        self.include_b = include & self.myfields.keys()
        self.builtins = builtins

        """ If builtins includes exc_info, the json encoder class must support TracebackException objects.
            if cls and default are both None, `genutility.json.TracebackExceptionEncoder` will be used.
        """

        if cls is None and default is None and "exc_info" in builtins:
            cls = TracebackExceptionEncoder
        self.dumps = partial(json.dumps, ensure_ascii=False, cls=cls, indent=None, default=default, sort_keys=sort_keys)

    def format(self, record: logging.LogRecord) -> str:
        # add builtins logger fields
        row = {name: getattr(record, name) for name in self.builtins}

        if "exc_info" in row:
            exc_info = row["exc_info"]
            if exc_info is not None:
                row["exc_info"] = TracebackException(*exc_info)

        # add custom logger fields
        if self.include_b:
            row.update({k: self.myfields[k]() for k in self.include_b})

        # add message fields
        row.update(record.msg)

        return self.dumps(row)



from builtins import sorted as _builtin_sorted
from copy import copy
from json import dumps
from typing import Any, List


def cast(obj, class_, instanceof=object, *args, **kwargs):
    """Changes the class of `obj` to `class_` if `obj` is an instance of `instanceof`,
    calls the initializer and returns it.
    """

    obj = copy(obj)
    if isinstance(obj, instanceof):
        obj.__class__ = class_
        obj.__init__(*args, **kwargs)
    else:
        raise TypeError(f"Object is not an instance of {instanceof.__name__}")
    return obj


class STAR:
    pass


def args_to_key(args: tuple, kwargs: dict, separator: Any = STAR) -> tuple:
    """Create cache key from function arguments."""

    key: List[tuple] = []
    if args:
        key.extend(args)
    if kwargs:
        key.append(separator)
        key.extend(sorted(kwargs.items()))

    return tuple(key)


def compress(value: Any) -> Any:
    """Creates a copy of the object where some data structures are replaced with equivalent ones
    which take up less space, but are not necessarily mutable anymore.

    tuple < list
    set == frozenset
    bytes < bytearray
    """

    # sets are not processed because they cannot contain lists or bytearrays anyway.

    if isinstance(value, (tuple, list)):  # tuple *can* contain mutables
        return tuple(compress(x) for x in value)
    elif isinstance(value, bytearray):
        return bytes(value)  # bytearray can only be bytes or List[int] right?
    elif isinstance(value, dict):
        return {k: compress(v) for k, v in value.items()}
    else:
        return value


def _sorted(value: list, *, reverse: bool = False) -> Any:
    def keyfunc(value: Any) -> Any:
        if isinstance(value, dict):
            return dumps(value)
        else:
            return value

    return _builtin_sorted(value, key=keyfunc)


def sorted(value: Any, *, reverse: bool = False) -> Any:
    if isinstance(value, list):
        return _sorted([sorted(v, reverse=reverse) for v in value], reverse=reverse)
    elif isinstance(value, tuple):
        return tuple(_sorted([sorted(v, reverse=reverse) for v in value], reverse=reverse))
    elif isinstance(value, dict):
        return dict(_sorted([(k, sorted(v, reverse=reverse)) for k, v in value.items()], reverse=reverse))
    else:
        return value


if __name__ == "__main__":
    from argparse import ArgumentParser
    from pathlib import Path

    from genutility.json import read_json, write_json

    parser = ArgumentParser(description="Sort json files recursively, both lists and objects")
    parser.add_argument("--inpath", type=Path)
    parser.add_argument("--outpath", type=Path)
    parser.add_argument("--indent")
    args = parser.parse_args()

    obj = read_json(args.inpath)
    write_json(sorted(obj), args.outpath, indent=args.indent)



import logging
import os
import signal
from ctypes import byref, c_wchar_p, cast, create_unicode_buffer, memset, sizeof
from ctypes.wintypes import DWORD, HANDLE, LPCWSTR, ULARGE_INTEGER
from enum import IntFlag
from msvcrt import get_osfhandle
from typing import IO, Union

from cwinsdk.shared.ehstorioctl import MAX_PATH
from cwinsdk.shared.ntdef import PWSTR
from cwinsdk.um import fileapi, winnt
from cwinsdk.um.combaseapi import CoTaskMemFree
from cwinsdk.um.consoleapi import ENABLE_VIRTUAL_TERMINAL_PROCESSING, GetConsoleMode, SetConsoleMode
from cwinsdk.um.knownfolders import FOLDERID_LocalAppData, FOLDERID_RoamingAppData
from cwinsdk.um.minwinbase import LOCKFILE_EXCLUSIVE_LOCK, LOCKFILE_FAIL_IMMEDIATELY, OVERLAPPED
from cwinsdk.um.processenv import GetStdHandle
from cwinsdk.um.shlobj_core import SHGetKnownFolderPath
from cwinsdk.um.winbase import STD_OUTPUT_HANDLE

from .os_shared import _usagetuple, _volumeinfotuple

PathType = Union[str, os.PathLike]


class FileAttributes(IntFlag):
    READONLY = winnt.FILE_ATTRIBUTE_READONLY
    HIDDEN = winnt.FILE_ATTRIBUTE_HIDDEN
    SYSTEM = winnt.FILE_ATTRIBUTE_SYSTEM
    DIRECTORY = winnt.FILE_ATTRIBUTE_DIRECTORY
    ARCHIVE = winnt.FILE_ATTRIBUTE_ARCHIVE
    DEVICE = winnt.FILE_ATTRIBUTE_DEVICE
    NORMAL = winnt.FILE_ATTRIBUTE_NORMAL
    TEMPORARY = winnt.FILE_ATTRIBUTE_TEMPORARY
    SPARSE_FILE = winnt.FILE_ATTRIBUTE_SPARSE_FILE
    REPARSE_POINT = winnt.FILE_ATTRIBUTE_REPARSE_POINT
    COMPRESSED = winnt.FILE_ATTRIBUTE_COMPRESSED
    OFFLINE = winnt.FILE_ATTRIBUTE_OFFLINE
    NOT_CONTENT_INDEXED = winnt.FILE_ATTRIBUTE_NOT_CONTENT_INDEXED
    ENCRYPTED = winnt.FILE_ATTRIBUTE_ENCRYPTED
    INTEGRITY_STREAM = winnt.FILE_ATTRIBUTE_INTEGRITY_STREAM
    VIRTUAL = winnt.FILE_ATTRIBUTE_VIRTUAL
    NO_SCRUB_DATA = winnt.FILE_ATTRIBUTE_NO_SCRUB_DATA
    EA = winnt.FILE_ATTRIBUTE_EA
    PINNED = winnt.FILE_ATTRIBUTE_PINNED
    UNPINNED = winnt.FILE_ATTRIBUTE_UNPINNED
    RECALL_ON_OPEN = winnt.FILE_ATTRIBUTE_RECALL_ON_OPEN
    RECALL_ON_DATA_ACCESS = winnt.FILE_ATTRIBUTE_RECALL_ON_DATA_ACCESS


def get_stdout_handle() -> HANDLE:
    """Might return a redirect handle. For the real handle, use CreateFile("CONOUT$")"""

    return GetStdHandle(STD_OUTPUT_HANDLE)


class EnableAnsi:  # doesn't work for some reason...
    def __init__(self) -> None:
        # os.system("") # calls cmd, which sets ANSI mode, but doesn't disable it when exiting (it's a bug probably)
        self.handle = get_stdout_handle()
        self.oldmode = DWORD()

        GetConsoleMode(self.handle, byref(self.oldmode))
        SetConsoleMode(self.handle, self.oldmode.value | ENABLE_VIRTUAL_TERMINAL_PROCESSING)

    def close(self) -> None:
        SetConsoleMode(self.handle, self.oldmode.value)

    def __enter__(self) -> None:
        return None

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


def _file_attributes(path: str) -> FileAttributes:
    return FileAttributes(fileapi.GetFileAttributesW(path))


def _islink(path: PathType) -> bool:
    """Tests if `path` refers to a symlink or a junction.
    - Python >= 3.2 `os.path.islink()` only supports symlinks, not junctions.
    - Python < 3.2 `os.path.islink()` always returns `False` on Windows.
    This function works in all cases.
    """

    # this can be replaced with a os.stat() based implementation for Python 3.8+

    filename = os.fspath(path)
    attributes = fileapi.GetFileAttributesW(filename)

    return attributes & winnt.FILE_ATTRIBUTE_REPARSE_POINT == winnt.FILE_ATTRIBUTE_REPARSE_POINT


def _realpath(path: PathType) -> str:
    """Partial fix for os.path.realpath() which doesn't work under Win7 for python < 3.8.
    This fails when it needs to concat a absolute unc path to a relative path. normpath doesn't normalize it then.
    """

    prefix = "\\\\?\\"
    _path = os.fspath(path)
    if _path.startswith(prefix):
        _path = _path[4:]
        was_dos_device_path = True
    else:
        was_dos_device_path = False
    _path = os.path.abspath(_path)
    if _islink(_path):
        _path = os.path.normpath(os.path.join(os.path.dirname(_path), os.readlink(_path)))
    if was_dos_device_path:
        _path = prefix + _path
    return _path


def _get_mount_path(path: str) -> str:
    FileName = LPCWSTR(path)
    BufferLength = 50  # MSDN suggestion
    VolumeName = create_unicode_buffer(BufferLength)
    fileapi.GetVolumeNameForVolumeMountPointW(FileName, VolumeName, BufferLength)
    return VolumeName.value


def _lock(fp: IO, exclusive: bool = True, block: bool = False) -> None:
    fd = fp.fileno()
    handle = get_osfhandle(fd)

    flags = 0
    if exclusive:
        flags |= LOCKFILE_EXCLUSIVE_LOCK
    if not block:
        flags |= LOCKFILE_FAIL_IMMEDIATELY

    overlapped = OVERLAPPED()
    memset(byref(overlapped), 0, sizeof(overlapped))

    fileapi.LockFileEx(handle, flags, 0, 0xFFFFFFFF, 0xFFFFFFFF, overlapped)


def _unlock(fp: IO) -> None:
    fd = fp.fileno()
    handle = get_osfhandle(fd)

    overlapped = OVERLAPPED()
    memset(byref(overlapped), 0, sizeof(overlapped))

    fileapi.UnlockFileEx(handle, 0, 0xFFFFFFFF, 0xFFFFFFFF, overlapped)


def _get_appdata_dir(roaming: bool = False) -> str:
    if roaming:
        folderid = FOLDERID_RoamingAppData
    else:
        folderid = FOLDERID_LocalAppData

    KF_FLAG_CREATE = 0x00008000

    rfid = byref(folderid)
    Flags = KF_FLAG_CREATE
    Token = None
    Path = PWSTR()

    try:
        result = SHGetKnownFolderPath(rfid, Flags, Token, byref(Path))
    except OSError:
        logging.error(f"SHGetKnownFolderPath result: {result & 0xffffffff:X}")
        raise

    ret = cast(Path, c_wchar_p).value
    CoTaskMemFree(Path)
    assert ret
    return ret


def _disk_usage_windows(path: str) -> _usagetuple:
    DirectoryName = LPCWSTR(path)
    FreeBytesAvailableToCaller = ULARGE_INTEGER(0)  # user free
    TotalNumberOfBytes = ULARGE_INTEGER(0)  # user total
    TotalNumberOfFreeBytes = None  # total free

    fileapi.GetDiskFreeSpaceExW(
        DirectoryName, byref(FreeBytesAvailableToCaller), byref(TotalNumberOfBytes), TotalNumberOfFreeBytes
    )
    return _usagetuple(
        TotalNumberOfBytes.value,
        TotalNumberOfBytes.value - FreeBytesAvailableToCaller.value,
        FreeBytesAvailableToCaller.value,
    )


def _volume_info_windows(path: str) -> _volumeinfotuple:
    if not path.endswith("\\"):
        raise ValueError("X: usually doesn't work. X:\\ does.")

    RootPathName = LPCWSTR(path)
    VolumeNameBuffer = create_unicode_buffer(MAX_PATH + 1)
    VolumeNameSize = MAX_PATH + 1
    VolumeSerialNumber = DWORD()
    MaximumComponentLength = DWORD()
    FileSystemFlags = DWORD()
    FileSystemNameBuffer = create_unicode_buffer(MAX_PATH + 1)
    FileSystemNameSize = MAX_PATH + 1

    fileapi.GetVolumeInformationW(
        RootPathName,
        VolumeNameBuffer,
        VolumeNameSize,
        byref(VolumeSerialNumber),
        byref(MaximumComponentLength),
        byref(FileSystemFlags),
        FileSystemNameBuffer,
        FileSystemNameSize,
    )

    return _volumeinfotuple(
        VolumeNameBuffer.value,
        VolumeSerialNumber.value,
        MaximumComponentLength.value,
        FileSystemFlags.value,
        FileSystemNameBuffer.value,
    )


def _interrupt_windows() -> None:
    os.kill(os.getpid(), signal.CTRL_C_EVENT)  # fixme: verify: works on win 10 but not on win 7


def _filemanager_cmd_windows(path: str) -> str:
    return f'explorer.exe /select,"{path}"'


if __name__ == "__main__":
    s = "\033[35m" + "color-test" + "\033[39m" + " test end"
    print(s)
    with EnableAnsi():
        print(s)
    print(s)



from typing import Iterable, Iterator, Union

from .exceptions import assert_choice


def encode_binary(boolit: Union[str, Iterator[bool]], pad: str = "0") -> bytes:
    """Encode a string consisting of 0s and 1s or an iterable returns boolean values
    to bytes.

    Example:
    encode_binary("1111111100000001") -> b'\xff\x01'
    encode_binary([True]*8+[False]*7+[True]) -> b'\xff\x01'
    """

    assert_choice("pad", pad, {"0", "1"})

    if isinstance(boolit, str):
        bin = boolit
    else:
        bin = "".join("1" if b else "0" for b in boolit)

    bin += pad * ((8 - len(bin) % 8) % 8)  # pad
    assert len(bin) % 8 == 0
    return bytes(int(bin[x : x + 8], 2) for x in range(0, len(bin), 8))


def _str2bool_it(s: Iterable[str]) -> Iterator[bool]:
    for chunk in s:
        for c in chunk:
            yield True if c == "1" else False


def decode_binary(key: bytes, tostring: bool = False) -> Union[str, Iterator[bool]]:
    """Decode bytes to either a string of 0s and 1s or an iterable of booleans."""

    it = (f"{b:08b}" for b in key)

    if tostring:
        return "".join(it)
    else:
        return _str2bool_it(it)



from collections import defaultdict
from typing import DefaultDict, Dict, Generic, Iterable, Optional, Set, Tuple, TypeVar, Union

T = TypeVar("T")
U = TypeVar("U")

_sentinel = object()


class CaseDict(Generic[T]):
    """Saves values for cased strings, but makes them retrievable using the lower cased version
    of the string as well.
    """

    def __init__(self) -> None:
        self.d: Dict[str, T] = {}
        self.casemap: DefaultDict[str, Set[str]] = defaultdict(set)

    def __len__(self) -> int:
        return len(self.d)

    def __setitem__(self, k: str, v: T) -> None:
        self.d[k] = v
        self.casemap[k.lower()].add(k)

    def keys(self) -> Iterable[str]:
        return self.d.keys()

    def ikeys(self) -> Iterable[str]:
        return self.casemap.keys()

    def values(self) -> Iterable[T]:
        return self.d.values()

    def items(self) -> Iterable[Tuple[str, T]]:
        return self.d.items()

    def __getitem__(self, k: str) -> T:
        return self.d[k]

    def igetitem(self, k: str) -> Dict[str, T]:
        results = self.casemap.get(k.lower())
        if results:
            return {k: self.d[k] for k in results}

        raise KeyError(k)

    def __contains__(self, k: str) -> bool:
        return k in self.d

    def icontains(self, k: str) -> bool:
        return k.lower() in self.casemap

    def __delitem__(self, k: str) -> None:
        del self.d[k]
        s = self.casemap[k.lower()]
        s.remove(k)
        if not s:
            del self.casemap[k.lower()]

    def idelitem(self, k: str) -> None:
        results = self.casemap.pop(k.lower())  # raise KeyError if it doesn't exist

        for r in results:
            del self.d[r]

    def get(self, k: str, default: Optional[T] = None) -> Optional[T]:
        return self.d.get(k, default)

    def iget(self, k: str, default: U = _sentinel) -> Union[Dict[str, T], U]:
        try:
            return self.igetitem(k)
        except KeyError:
            if default is _sentinel:
                return {}
            else:
                return default

    # new

    def igetitemset(self, k: str) -> Set[str]:
        return self.casemap[k.lower()]

    def igetset(self, k: str, default: U = _sentinel) -> Union[Set[str], U]:
        if default is _sentinel:
            default = set()
        return self.casemap.get(k.lower(), default)
